//=============================================================================
//
//      cpuio.c
//
//      CPU/Board Specific IO
//
// Author(s):    Michael Kelly, Cogent Computer Systems, Inc.
// Date:         05-26-2002
// Description:  This file contains the IO functions required by Micro Monitor
//				 that are unique to each CPU/Board combination
//
//
//=============================================================================

#include "config.h"
#include "cpuio.h"
#include "genlib.h"
#include "ether.h"
#include "stddefs.h"
#include "uart16550.h"
#include "cpu_pci.h"
#include "ps2_kbd.h"
#include "au1500_gpio.h"
#include "warmstart.h"

extern int fs6377_init(int mode);
extern int ps2_kbd_init(void);
extern int ps2_gotachar(void);
extern uchar ps2_getchar(void); 
extern void sed_pwr_dn(void);
extern int sed135x_init(void);
extern void sed_lcd_bkl(uchar bright);
extern void sed_putchar(char c);
extern void pci_init();
extern void GPIO_init();
extern int GPIO_clr(int);
extern int GPIO_set(int);
extern int pci_probe_bus(int bus);
extern void sed135x_on(void);
extern void tlbwrite(int, ulong);

// globals
extern int sed_color_depth;
extern int sed135x_ok;
extern int next_bus;
extern int mem_base;
extern int io_base;
extern int pci_dev;

// structure to setup the MMU TLB
typedef struct
{
    unsigned long   EntryLo0;
    unsigned long   PageMask;
    unsigned long   EntryLo1;
    unsigned long   EntryHi;
} TLBEntry;

// Delay for some usecs. - Not accurate, assumes ROM mode
// and no Cache
void udelay(int delay)
{
  int i;
  for ( i = LOOPS_PER_USEC * delay; i ; i--)
	;
}

// We use the SED GPIO1 to control the LCD backlight
void
lcd_bkl(uchar bright)
{
#if INCLUDE_DISPLAY
	sed_lcd_bkl(bright);
#endif
}


/* InitUART():
 * This is the top-level UART initialization for the monitor. 
 */
void
InitUART(int baud)
{
	/* set the failsafe serial port to a sane state */
	// enable the DUART
	UARTA(UART_ER) = UART_ER_CE;	// enable the clocks first
	UARTA(UART_ER) = UART_ER_CE | UART_ER_ME;	// now enable the module

	// Au1500 uses a dedicated clock divsor register.
	// clocks are derived from MEMCLK/2.
	UARTA(UART_CDR) = (((MEMCLK/2)/16)/baud);

	UARTA(UART_LCTL) = 0x03;	/* DLAB off, 8-1-N */
	UARTA(UART_MCTL) = 0x01;	/* drop all lines - raise DTR */
	UARTA(UART_IEN) = 0x00;	/* disable all interrupts */
	UARTA(UART_FCTL) = 0x00;	/* turn on FIFOs & clear then */
}

int
ConsoleBaudSet(int baud)
{
	// Au1500 uses a dedicated clock divsor register.
	// clocks are derived from MEMCLK/2.
	UARTA(UART_CDR) = (((MEMCLK/2)/16)/baud);
	return(0);
}

int
devInit(int baud)
{

	// initialize the GPIO ports
	GPIO_init();

	// Initialize the UART first so we can print messages
    InitUART(baud);

	// Initialize the display controller
#if INCLUDE_DISPLAY
	sed_color_depth = 4;
	sed135x_init();
	if (sed135x_ok == 0)
		printf("sed135x_init() failed!\n");
	sed135x_on();
#endif

	// Initialize the PS2 controller and keyboard - if attached
#if INCLUDE_PS2KBD
	ps2_kbd_init();
#endif

	// Probe the pci bus for devices and fill in pci_dev_info
//	pci_init();
//	next_bus = 0;
//	mem_base = PCI_MEM_BASE;
//	io_base = PCI_IO_BASE;
//	pci_dev = 0;
//	pci_probe_bus(0);

	GPIO_clr(GPIO_USER_LED0);
	GPIO_set(GPIO_USER_LED1);
	GPIO_set(GPIO_USER_LED2);

	return 0;
}

int
target_console_empty(void)
{
	if (!(UARTA(UART_LSTAT) & SIO_LSTAT_TRDY))
		return(1);
	return(0);
}

int
target_putchar(char c)
{
	ulong temp32;

    /* Wait for transmit ready bit */
    while(1) {
        if (UARTA(UART_LSTAT) & SIO_LSTAT_TRDY)
            break;
    }

	temp32 = c & 0xff;
	UARTA(UART_TXD) = temp32;

#if INCLUDE_DISPLAY
	sed_putchar(c);
#endif

    return((int)c);
}

int
target_gotachar(void)
{
    if (UARTA(UART_LSTAT) & SIO_LSTAT_RRDY)
        return(1);
#if INCLUDE_PS2KBD
	if(ps2_gotachar())
       return(1);
#endif
    return(0);
}

int 
target_getchar(void)
{
    char    c;

#if INCLUDE_PS2KBD
	if(ps2_gotachar())
		c = ps2_getchar();
	else
#endif
	    c = (uchar)UARTA(UART_RXD);
    return((int)c);
}

void
initCPUio()
{
    TLBEntry        tlb;

	// setup the TLB
	// Do the S1D13506 first
    tlb.PageMask = 0x01ffe000;			// pagemask for 16M pages
    tlb.EntryHi = SED_REG_BASE;   		// Virtual Address
    tlb.EntryLo0 = (0x38000017			// 0xE space, CCA = 2, D, V and G set
		 | (SED_REG_BASE >> 6));		// Physical Address
    tlb.EntryLo1 = tlb.EntryLo0;

	tlbwrite(0,(ulong)&tlb);
	
	// Now do the HT6542B
    tlb.PageMask = 0x01ffe000;			// pagemask for 16M pages
    tlb.EntryHi = PS2_BASE;				// Virtual Address
    tlb.EntryLo0 = (0x34000017			// 0xD space, CCA = 2, D, V and G set
    			 | (PS2_BASE >> 6));	// Physical Address
    tlb.EntryLo1 = tlb.EntryLo0;

	tlbwrite(1,(ulong)&tlb);
	
	return;

}

void
intsrestore(ulong msr)
{
}

ulong
intsoff(void)
{
	return(0);
}

void
cacheInitForTarget(void)
{
}

void
target_reset(void)
{
	monrestart(INITIALIZE);
}


/* Stubs until we establish more exception handling...
 */
void
exc080(void)
{
	puts("EXCEPTION 080!\n");
	while(1);
}
void
exc100(void)
{
	puts("EXCEPTION 100!\n");
	while(1);
}
void
exc180(void)
{
	puts("EXCEPTION 180!\n");
	while(1);
}
void
exc200(void)
{
	puts("EXCEPTION 200!\n");
	while(1);
}
void
exc280(void)
{
	puts("EXCEPTION 280!\n");
	while(1);
}
void
exc300(void)
{
	puts("EXCEPTION 300!\n");
	while(1);
}
void
exc380(void)
{
	puts("EXCEPTION 380!\n");
	while(1);
}
void
exc400(void)
{
	puts("EXCEPTION 400!\n");
	while(1);
}
void
exc480(void)
{
	puts("EXCEPTION 480!\n");
	while(1);
}

