//==========================================================================
//
//      cpu_pci.c
//
//      CPU Specific PCI support and initialization routines
//
// Author(s):    Michael Kelly, Cogent Computer Systems, Inc.
// Contributors: 
// Date:         02/05/2002
// Description:  This file contains PCI intialization and support routines
//               specific to the MPC8245 as implemented on the CSB281 
//
//--------------------------------------------------------------------------

#include "config.h"
#include "cpuio.h"
#include "stddefs.h"
#include "genlib.h"
#include "pci.h"
#include "cpu_pci.h"

//--------------------------------------------------------------------------
// Function Prototypes
//
void set_cfg_type_id(ulong bus, ulong dev_func);
ulong pci_cfg_read_dword (ulong bus, ulong dev_func, ulong offset);
ushort pci_cfg_read_word (ulong bus, ulong dev_func, ulong offset);
uchar pci_cfg_read_byte (ulong bus, ulong dev_func, ulong offset);
void pci_cfg_write_dword (ulong bus, ulong dev_func, ulong offset, ulong data);
void pci_cfg_write_word (ulong bus, ulong dev_func, ulong offset, ushort data);
void pci_cfg_write_byte (ulong bus, ulong dev_func, ulong offset, uchar  data);
int pci_master_abort(void);
void pci_init(void);

// external functions
extern void udelay(int delay);

// external data
extern ulong sh_mem_base;

//--------------------------------------------------------------------------
// PCI Support routines - specific to MPC8245 Bridge
//

//--------------------------------------------------------------------------
// Config Space read/write routines - CPU specific files must define:
// PCI_CFG_ADD, PCI_CFG_DATA_32, PCI_CFG_DATA_16 and PCI_CFG_DATA_8.  The
// support routines set_cfg_type_id() and pci_master_abort() are also CPU
// specific.
// 
ulong pci_cfg_read_dword (ulong bus, ulong dev_func, ulong offset)
{
    ulong config_dword;

	// select the device and make sure we have the right config type
	set_cfg_type_id(bus, dev_func);

#ifdef PCI_DBG
	printf("pci_cfg_read_dword() Bus: %d, Dev: %d, Func: %d, ID: %02x\n", 
								 bus, PCI_GET_DEV(dev_func), PCI_GET_FUNC(dev_func), BCSR & 0x7);
#endif
	PCI_CFG_ADD = HTOLE32(PCI_CFG_EN 		
					| PCI_CFG_BUS(bus)	
					| PCI_CFG_DEV(PCI_GET_DEV(dev_func)) 
					| PCI_CFG_FUNC(PCI_GET_FUNC(dev_func))
					| PCI_CFG_REG(offset));

	config_dword = HTOLE32(PCI_CFG_DATA_32(offset));
	asm volatile ("eieio");

#ifdef PCI_DBG
	printf("pci_cfg_read_dword() CFG_ADD: %08lx, CFD_DATA: %08lx\n", PCI_CFG_ADD, config_dword);
#endif

//    if (pci_master_abort()) return 0xffffffff;

    return config_dword;
}

ushort pci_cfg_read_word (ulong bus, ulong dev_func, ulong offset)
{
    ushort config_word;

	set_cfg_type_id(bus, dev_func);

#ifdef PCI_DBG
	printf("pci_cfg_read_word() Bus: %d, Dev: %d, Func: %d, ID: %02x\n", 
								bus, PCI_GET_DEV(dev_func), PCI_GET_FUNC(dev_func), BCSR & 0x7);
#endif
	PCI_CFG_ADD = HTOLE32(PCI_CFG_EN 		
					| PCI_CFG_BUS(bus)	
					| PCI_CFG_DEV(PCI_GET_DEV(dev_func)) 
					| PCI_CFG_FUNC(PCI_GET_FUNC(dev_func))
					| PCI_CFG_REG(offset));

	config_word = HTOLE16(PCI_CFG_DATA_16(offset));
	asm volatile ("eieio");

#ifdef PCI_DBG
	printf("pci_cfg_read_word() CFG_ADD: %08lx, CFD_DATA: %04x\n", PCI_CFG_ADD, config_word);
#endif
//    if (pci_master_abort()) return 0xffff;

    return config_word;
}

uchar pci_cfg_read_byte (ulong bus, ulong dev_func, ulong offset)
{
    uchar config_byte;

	set_cfg_type_id(bus, dev_func);

#ifdef PCI_DBG
	printf("pci_cfg_read_byte() Bus: %d, Dev: %d, Func: %d, ID: %02x\n", 
								bus, PCI_GET_DEV(dev_func), PCI_GET_FUNC(dev_func), BCSR & 0x7);
#endif
	PCI_CFG_ADD = HTOLE32(PCI_CFG_EN 		
					| PCI_CFG_BUS(bus)	
					| PCI_CFG_DEV(PCI_GET_DEV(dev_func)) 
					| PCI_CFG_FUNC(PCI_GET_FUNC(dev_func))
					| PCI_CFG_REG(offset));

	config_byte = PCI_CFG_DATA_8(offset);
	asm volatile ("eieio");

#ifdef PCI_DBG
	printf("pci_cfg_read_byte() CFG_ADD: %08lx, CFD_DATA: %02x\n", PCI_CFG_ADD, config_byte);
#endif
//    if (pci_master_abort()) return 0xff;

    return config_byte;
}

void pci_cfg_write_dword (ulong bus, ulong dev_func, ulong offset, ulong data)
{

	set_cfg_type_id(bus, dev_func);

#ifdef PCI_DBG
	printf("pci_cfg_write_dword() Bus: %d, Dev: %d, Func: %d, ID: %02x\n", 
								  bus, PCI_GET_DEV(dev_func), PCI_GET_FUNC(dev_func), BCSR & 0x7);
#endif
	PCI_CFG_ADD = HTOLE32(PCI_CFG_EN 		
					| PCI_CFG_BUS(bus)	
					| PCI_CFG_DEV(PCI_GET_DEV(dev_func)) 
					| PCI_CFG_FUNC(PCI_GET_FUNC(dev_func))
					| PCI_CFG_REG(offset));

    PCI_CFG_DATA_32(offset) =  HTOLE32(data);
	asm volatile ("eieio");

#ifdef PCI_DBG
	printf("pci_cfg_write_dword() CFG_ADD: %08lx, CFD_DATA: %08lx\n", PCI_CFG_ADD, data);
#endif
}

void pci_cfg_write_word (ulong bus, ulong dev_func, ulong offset, ushort data)
{
	set_cfg_type_id(bus, dev_func);

#ifdef PCI_DBG
	printf("pci_cfg_write_word() Bus: %d, Dev: %d, Func: %d, ID: %02x\n", 
	                             bus, PCI_GET_DEV(dev_func), PCI_GET_FUNC(dev_func), BCSR & 0x7);
#endif
	PCI_CFG_ADD = HTOLE32(PCI_CFG_EN 		
					| PCI_CFG_BUS(bus)	
					| PCI_CFG_DEV(PCI_GET_DEV(dev_func)) 
					| PCI_CFG_FUNC(PCI_GET_FUNC(dev_func))
					| PCI_CFG_REG(offset));

    PCI_CFG_DATA_16(offset) =  HTOLE16(data);
	asm volatile ("eieio");

#ifdef PCI_DBG
	printf("pci_cfg_write_word() CFG_ADD: %08lx, CFD_DATA: %04x\n", PCI_CFG_ADD, data);
#endif
}

void pci_cfg_write_byte (ulong bus, ulong dev_func, ulong offset, uchar  data)
{
	set_cfg_type_id(bus, dev_func);

#ifdef PCI_DBG
	printf("pci_cfg_write_byte() Bus: %d, Dev: %d, Func: %d, ID: %02x\n", 
								 bus, PCI_GET_DEV(dev_func), PCI_GET_FUNC(dev_func), BCSR & 0x7);
#endif
	PCI_CFG_ADD = HTOLE32(PCI_CFG_EN 		
					| PCI_CFG_BUS(bus)	
					| PCI_CFG_DEV(PCI_GET_DEV(dev_func)) 
					| PCI_CFG_FUNC(PCI_GET_FUNC(dev_func))
					| PCI_CFG_REG(offset));

    PCI_CFG_DATA_8(offset) =  data;
	asm volatile ("eieio");

#ifdef PCI_DBG
	printf("pci_cfg_write_byte() CFG_ADD: %08lx, CFD_DATA: %02x\n", PCI_CFG_ADD, data);
#endif
}

//--------------------------------------------------------------------------
// set_cfg_type_id()
//
// This support function sets the ID bit and the correct config type
// It is used by the various pci_cfg_read_xx and write_xx
// functions that follow.
void set_cfg_type_id(ulong bus, ulong dev_func){

	BCSR = BCSR & ~(IDSEL1 | IDSEL2 | IDSEL3);	// 1st clear all idsel bits.
	// if bus is non-zero, it means we have to do a type 1 config cycle
	// to reach through the device (which must be a bridge) to the
	// other bus.  The actual bus number does not matter since any non-zero
	// bus by definition is on the other side of a bridge.
	if (bus == 0) {
		// Bus 0
		// convert a device number into a single bit in the ID Select register
		if(PCI_GET_DEV(dev_func)){
			switch(PCI_GET_DEV(dev_func))
			{
				case DEV_START + 0 : BCSR = BCSR | IDSEL1; break;
				case DEV_START + 1 : BCSR = BCSR | IDSEL2; break;
				case DEV_START + 2 : BCSR = BCSR | IDSEL3; break;
				default : break;
			}
		}
	}
}

//--------------------------------------------------------------------------
// pci_master_abort()
//
// This support function returns -1 if the Master Abort bit in the PCI 
// status register is set.  It is bridge dependant in case any special
// steps need to be taken to recover from a master abort.  Most bridge's
// just need a one written to the master abort bit to clear it.
int pci_master_abort(){

	ushort temp16;

	PCI_CFG_ADD = HTOLE32(PCI_CFG_EN | PCI_CFG_STAT_REG);
#ifdef PCI_DBG
	printf("Check Master Abort, CFG_ADD: %08lx, ", PCI_CFG_ADD);
#endif
	temp16 = HTOLE16(PCI_CFG_DATA_16(PCI_CFG_STAT_REG));
#ifdef PCI_DBG
	printf("CFG_DATA: %04x\n", temp16);
#endif
	// clear the abort bit by writing one to it
	if (temp16 & PCI_CFG_STAT_MST_ABRT){
		PCI_CFG_DATA_16(PCI_CFG_STAT_REG) = PCI_CFG_STAT_MST_ABRT;	
#ifdef PCI_DBG
		printf("Master Abort!\n");
#endif
		return -1;
	}

	// no master abort occured
	return 0;	

}

//-------------------------------------------------------------------------
// Initialize the bridge
//
void pci_init(void)
{
	printf("Initializing the MPC8245 PCI Bridge\n");

	// PCI windows are fixed so no setup is required

	// Enable BAR0 and set PCI Side window address to 0.
	pci_cfg_write_dword (0, 0, PCI_CFG_BAR0_REG, 0x01);

	// enable 8245 for memory accesses and mastering
	pci_cfg_write_word (0, 0, PCI_CFG_CMD_REG, (PCI_CFG_CMD_MEM | PCI_CFG_CMD_MASTER));

	printf("MPC8245 PCI Initialization Complete\n");

	// initialize the shared memory base address
	sh_mem_base = PCI_SM_BASE;

}

//--------------------------------------------------------------------------
// End of cpu_pci.c
