    .file "rom_reset.s"

/*
 * General notice:
 * This code is part of a boot-monitor package developed as a generic base
 * platform for embedded system designs.  As such, it is likely to be
 * distributed to various projects beyond the control of the original
 * author.  Please notify the author of any enhancements made or bugs found
 * so that all may benefit from the changes.  In addition, notification back
 * to the author will allow the new user to pick up changes that may have
 * been made by other users after this version of the code was distributed.
 *
 * Author:  Ed Sutter
 * email:   esutter@lucent.com
 * phone:   908-582-2351
 * 
 *
 * Modified for the CSB226 - PXA250 Single Board
 *
 * Modified by Technology Rescue.
 *	12.03.2003 : changed frequency select.
 *	29.04.2003 : changed K0DB2 bit in MDREFR reg.
 *	30.10.2003 : changed memory bank size in MSC_MDCNFG reg.
 *	31.10.2003 : changed SDRAM addressing muxing mode to SA1111 mode.
 *	03.11.2003 : restored 100 MHz SDRAM bus speed.
 *
 * rom_reset.s:
  */

#include "warmstart.h"
#include "pxa250.h"
#include "pxa250_msc.h"
    /*
     * Have a separate stack for each processor mode.
     */

    /* define sizes for each mode's stack */
    .equ MonStackSz, 4096
    .equ FiqStackSz, 4096
    .equ IrqStackSz, 4096
    .equ AbtStackSz, 4096
    .equ UndStackSz, 4096
    .equ SysStackSz, 4096

    /* declare the stacks */
    .global MonStack
    .global FiqStack
    .global IrqStack
    .global AbtStack
    .global UndStack
    .global SysStack        
    
    /* allocate the stacks */
    .comm   MonStack, MonStackSz    /* for the SVC mode */
    .comm   FiqStack, FiqStackSz    /* for the FIQ mode */
    .comm   IrqStack, IrqStackSz    /* for the IRQ mode */
    .comm   AbtStack, AbtStackSz    /* for the Abort mode */
    .comm   UndStack, UndStackSz    /* for the Undef mode */
    .comm   SysStack, SysStackSz    /* for the System mode */
    /* User mode has the same stack as system mode. */

/*********************************************************************/
    
    .extern start

    .global reset
    .global coldstart
    .global lukewarmstart
    .global warmstart
    .global ipaddr
    .global etheraddr

    .text
        
    /*
     * Exception table at address 0
     */ 
reset:  
    b coldstart
    b undefined_instruction
    b software_interrupt
    b abort_prefetch
    b abort_data
    b not_used
    b interrupt_request
    b fast_interrupt_request

/********************************************************************/
    
    /*
     * Provide space to allow a programmer to place an ascii 
     * string in each of these locations as an optional point
     *  of storage for MAC and/or IP address...
     */ 

ipaddr:
    .byte 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
    .byte 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
    .byte 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
    
etheraddr:
    .byte 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
    .byte 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
    .byte 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff

undefined_instruction:
	b undefined_instruction

software_interrupt:
	b software_interrupt

abort_prefetch:
	b abort_prefetch

abort_data:
	b abort_data

not_used:
	b not_used

interrupt_request:
	b interrupt_request

fast_interrupt_request:
	b fast_interrupt_request

/*********************************************************************/
coldstart:

    /*
     * At the end of the reset sequence, MMU, Icache, Dcache,
     * and write buffer are all disabled.
     * Also IRQs and FIQs are disabled in the processor's CPSR
     * The operating mode is SVC (supervisory mode), and the
     * PC is vectored at 0x00000000. A branch in 0x00000000
     * brings us directly here.
	 *
	 */

	ldr r0, =0x2001					/* allow access to all coprocessors */
	mcr 	p15,0,r0,c15,c1,0
	nop
	nop
	nop

	ldr 	r0, =0x00000078
	mcr 	p15,0,r0,c1,c0,0		/* Disable MMU, caches, write buffer */
	nop
	nop
	nop

	ldr 	r0, =0x00000000
	mcr     p15,0,r0,c8,c7,0   		/* flush TLB's */
	mcr     p15,0,r0,c7,c7,0   		/* flush Caches */
	mcr		p15,0,r0,c7,c10,4		/* Flush Write Buffer */
	nop
	nop
	nop

	mvn    r0, #0                   /* grant manager access to all domains */
	mcr    p15,0,r0,c3,c0,0  

	/*
     * Initialize the PXA250 interrupt controller
     */

    /* Mask off all IRQs and FIQs */

    ldr r1, =(INT_BASE | INT_ICMR)
    ldr r2, =0x0	        	/* mask everything off */
    str r2, [r1]  			/* set mask (write to ICMR) */

    /*
     * Initialize PXA250 Clock Register
     */
    ldr r1, =CLK_BASE			/* base of clock registers */

@ddlc 120303    ldr r2, =0x00000141
    ldr r2, =0x00000241
    str r2, [r1, #CLK_CCCR]		/* you must read the MSC0/1 registers before they */      
	ldr r3, [r1, #CLK_CCCR]		/* will update and work properly */

@ddlc 120303 try turbo mode

/* Internal 32.768 kHz clock takes up to 10 sec to stabilize. */
/* Enable the 32Khz oscillator for RTC and PowerManager. */

	ldr	r1, =CLK_OSCC
	mov	r2, #0
	str	r2, [r1]

/* Change frequency now by setting bit 2 = FCS
/* FCS can be reset later to prevent undesired freq change. */

	mov	r1, #3							/* frequency change + turbo mode bits. */
	mcr	p14, 0, r1, c6, c0, 0       	/* write CCLKCFG. */
	nop
	nop
	nop
	nop
	nop
	nop
	nop

    /*
     * Initialize PXA250 Memory Controller
     */
    ldr r1, =MSC_BASE		/* base of memory controller */

@ddlc    ldr r2, =0x2ef15af0
    ldr r2, =0x2ef11dd8
    str r2, [r1, #MSC_MSC0]		/* you must read the MSC0/1 registers before they */      
	ldr r3, [r1, #MSC_MSC0]		/* will update and work properly */

    ldr r2, =0x00003ff4
    str r2, [r1, #MSC_MSC1]
    ldr r3, [r1, #MSC_MSC1]
    
    ldr r2, =0x03ca4fff			/* build MDREFR in a specific order */
    str r2, [r1, #MSC_MDREFR]
    
    ldr r2, =0x03ca4030			/* REF Rate = (64MS/4096 Rows)/32 = 48 */
    str r2, [r1, #MSC_MDREFR]
    
@ddlc29042003    ldr r2, =0x03cf6030			/* Set K0RUN, K1RUN and K2RUN */
    ldr r2, =0x03cd2030			/* Set K0RUN, K1RUN and K2RUN, set K0DB2, bit14 & bit 17 to 0 */
    str r2, [r1, #MSC_MDREFR]
    
	ldr r2, =0x038d2030			/* Clear Self Refresh */
    str r2, [r1, #MSC_MDREFR]
    
    ldr r2, =0x038db030			/* Set E0PIN and E1PIN */
    str r2, [r1, #MSC_MDREFR]
    
@ddlc30102003    ldr r2, =0x09a809a8			/* Set CNFG, but don't enable just yet */
    ldr r2, =0x09C819C8			/* Set CNFG, but don't enable just yet */
    str r2, [r1, #MSC_MDCNFG]

	/* pause for ~200usec for SDRAM to init */

	ldr		r3, =0x400			/* loop count */

do_delay:
	nop
	nop
	subs	r3,r3,#1
	bne		do_delay
	nop
	nop
	nop

	/* force 9 cycles to SDRAM */
	ldr     r2, =0xa0000000
	str     r2, [r2]
	str     r2, [r2]
	str     r2, [r2]
	str     r2, [r2]
	str     r2, [r2]
	str     r2, [r2]
	str     r2, [r2]
	str     r2, [r2]
	str     r2, [r2]
	
	/* now enable SDRAM */
@ddlc30102003    ldr r2, =0x09a909a9
    ldr r2, =0x09C919C9
    str r2, [r1, #MSC_MDCNFG]

	/* and do an MRS */
    ldr r2, =0x00220022
    str r2, [r1, #MSC_MDMRS]

/********************************************************************/

midstart:
    ldr r0, =INITIALIZE
    
    /* fall-through to 'lukewarmstart' */

/********************************************************************/
    
lukewarmstart:  
    /* Save the argument to r11 */
    mov r11, r0

    /*
     * *** DO NOT TOUCH R11 ***
     */ 

    /*
     * Set-up the stack-pointers for all operating modes
     */

    /* FIQ mode */
    mrs r0, cpsr                /* move CPSR to r0 */
    bic r0, r0, #0x1f           /* clear all mode bits */
    orr r0, r0, #0x11           /* set FIQ mode bits */
    msr CPSR_c, r0              /* move back to CPSR */
    ldr sp, =(FiqStack + FiqStackSz - 4)    /* initialize the stack ptr */
    /* IRQ mode */
    mrs r0, cpsr                /* move CPSR to r0 */
    bic r0, r0, #0x1f           /* clear all mode bits */
    orr r0, r0, #0x12           /* set IRQ mode bits */
    msr CPSR_c, r0              /* move back to CPSR */
    ldr sp, =(IrqStack + IrqStackSz - 4)    /* initialize the stack ptr */
    /* Abort mode */
    mrs r0, cpsr                /* move CPSR to r0 */
    bic r0, r0, #0x1f           /* clear all mode bits */
    orr r0, r0, #0x17           /* set Abort mode bits */
    msr CPSR_c, r0              /* move back to CPSR */
    ldr sp, =(AbtStack + AbtStackSz - 4)    /* initialize the stack ptr */
    /* Undef mode */
    mrs r0, cpsr                /* move CPSR to r0 */
    bic r0, r0, #0x1f           /* clear all mode bits */
    orr r0, r0, #0x1b           /* set Undef mode bits */
    msr CPSR_c, r0              /* move back to CPSR */
    ldr sp, =(UndStack + UndStackSz - 4)    /* initialize the stack ptr */
    /* System mode */
    mrs r0, cpsr                /* move CPSR to r0 */
    bic r0, r0, #0x1f           /* clear all mode bits */
    orr r0, r0, #0x1f           /* set System mode bits */
    msr CPSR_c, r0              /* move back to CPSR */
    ldr sp, =(SysStack + SysStackSz - 4)    /* initialize the stack ptr */
    /* 'warmstart' will take us back to SVC mode
       stack for SVC mode will also be setup in warmstart */

    mov r0, r11     /* get argument back from r11 */
    b   warmstart

    
/********************************************************************/

bailout:
    ldr r0, =BAILOUT
    b   warmstart

/********************************************************************/
    
warmstart:
    /* Save the argument to r11 */
    mov r11, r0

    /*
     * *** DO NOT TOUCH R11 ***
     */ 


    /* Change (back) to SVC mode */
    mrs r0, cpsr                /* move CPSR to r0 */
    bic r0, r0, #0x1f           /* clear all mode bits */
    orr r0, r0, #0x13           /* set System mode bits */
    msr CPSR_c, r0              /* move back to CPSR */
    /* Reset the stack pointer for the SVC mode (our current mode) */
    ldr sp, =(MonStack + MonStackSz - 4)

    /*
     * Restore argument which was saved to r11 and jump to
     * the C function start().
     */ 
    
    mov r0, r11
jump_to_c:
    bl cstart

    /* the C code should never return */
    b reset

.align 4

