//==========================================================================
//
// csb637_gpio.h
//
// Author(s):
//		Michael Kelly, Cogent Computer Systems, Inc.
// Modifications: 
//		10/15/2005 Updated to conform to uMon's generic GPIO structure
// Description:
//		This file contains the board level to vitual GPIO bit assignments
//
// The AT91RM9200 GPIO's are spread across four 32-bit ports A-D.
// For bits that have an alternate function assigned, we use the bit 
// position within the respective port as we need that to setup the
// registers.  But for bits that are used as GPIOs we assign a 
// virtual GPIO number which is used by the GPIO functions
//
// This section also defines the bit setting for the alternate
// functions of the pins.  If we want alternate function A, we prepend
// an A, for function B, a B.  This insures we will get an error, or
// at least it will be very visible if we try to assign a function
// to the wrong side.  Note the A or B is not related to the Port A
// B, just the alternate function.
//
// Finally we build the default values to be written into the GPIO registers
// in gpio_init (in csb637_gpio.c).

#include "umongpio.h"

#ifndef CSB637_GPIO_H
#define CSB637_GPIO_H

// PORT A						   BIT	A/B	DIR	SENSE	DESCRIPTION
#define PA0			GPIO(0,0)	// PA0  -	-	-	Expansion GPIO/MISO/PCK3
#define	PA1			GPIO(1,1)	// PA1  -	-	-	Expansion GPIO/MOSI/PCK0
#define PA2			GPIO(2,2)	// PA2  -	-	-	Expansion GPIO/SPCK/IRQ4
#define PA3			GPIO(3,3)	// PA3  -	-	-	Expansion GPIO/NPCS0/IRQ5
#define PA4			GPIO(4,4)	// PA4  -	-	-	Expansion GPIO/NPCS1/PCK1
#define PA5			GPIO(5,5)	// PA5  -	-	-	Expansion GPIO/NPCS2/TXD3
#define PA6			GPIO(6,6)	// PA6  - 	-	-	Expansion GPIO/NPCS3/RXD3
#define A_ETXCK		GPIO(7,7)	// PA7  A	-	-	Ethernet Transmit Clock
#define A_ETXEN		GPIO(8,8)	// PA8  A	-	-	Ethernet Transmit Enable
#define A_ETXD0		GPIO(9,9)	// PA9  A	-	-	Ethernet Transmit Data 0
#define A_ETXD1		GPIO(10,10)	// PA10 A	-	-	Ethernet Transmit Data 1
#define A_ECRS		GPIO(11,11)	// PA11 A	-	-	Ethernet Carrier Sense 
#define A_ERXD0		GPIO(12,12)	// PA12 A	-	-	Ethernet Receive Data 0
#define A_ERXD1		GPIO(13,13)	// PA13 A	-	-	Ethernet Receive Data 1
#define A_ERXER		GPIO(14,14)	// PA14 A	-	-	Ethernet Receive Error
#define A_EMDC		GPIO(15,15)	// PA15 A	-	-	Ethernet MD Clock
#define A_EMDIO		GPIO(16,16)	// PA16 A	-	-	Ethernet MD Data
#define PA17		GPIO(17,17)	// PA17 - 	In	-	Expansion GPIO/TXD0/TIOA0
#define PA18		GPIO(18,18)	// PA18 - 	In	-	Expansion GPIO/RXD0/TIOB0
#define LCD_BKL		GPIO(19,19)	// PA19 - 	Out	Hi	LCD Backlight, 1 = on
#define CF_IRQ		GPIO(20,20)	// PA20 - 	In	Hi	Compact Flash IRQ
#define LCD_VEE		GPIO(21,21)	// PA21 - 	In	Hi	LCD VEE, 1 = on
#define PA22		GPIO(22,22)	// PA22 - 	-	-	Expansion GPIO/RXD2/TIOB2
#define PA23		GPIO(23,23)	// PA23 - 	-	-	Expansion GPIO/TXD2/IRQ3
#define CRT_DET		GPIO(24,24)	// PA24 - 	In	Lo	CRT Detect
#define SDA			GPIO(25,25)	// PA25 - 	I/O	-	I2C Data - GPIO mode
#define SCL			GPIO(26,26)	// PA26 - 	I/O	-	I2C Clock - GPIO mode
#define A_SDCK		GPIO(27,27)	// PA27 A 	- 	- 	SD Card Clock 
#define A_SDCMD		GPIO(28,28)	// PA28 A 	- 	- 	SD Card Command
#define A_SDD0		GPIO(29,29)	// PA29 A 	- 	- 	SD Card Data 0
#define A_DBG_RXD	GPIO(30,30)	// PA30 A 	- 	- 	Debug Receive
#define A_DBG_TXD	GPIO(31,31)	// PA31 A 	- 	-	Debug Transmit

// default settings for PORT A
// Peripheral A select - SPI, SD Card 0, Ethernet and Debug UART
#define DEF_PIOA_ASR	\
	(GPIOBIT(A_ETXCK) | GPIOBIT(A_ETXEN) | GPIOBIT(A_ETXD0) | \
	GPIOBIT(A_ETXD1) | GPIOBIT(A_ECRS) | GPIOBIT(A_ERXD0) | \
	GPIOBIT(A_ERXD1) | GPIOBIT(A_ERXER) | GPIOBIT(A_EMDC) | \
	GPIOBIT(A_EMDIO) | GPIOBIT(A_SDCK) | GPIOBIT(A_SDCMD) | \
	GPIOBIT(A_SDD0) | GPIOBIT(A_DBG_RXD) | GPIOBIT(A_DBG_TXD))

// Peripheral B select - None
#define DEF_PIOA_BSR	0	

 // enable pullups where desired - usually just expansion GPIO
#define DEF_PIOA_PUER	\
	(GPIOBIT(PA0) | GPIOBIT(PA1) | GPIOBIT(PA2) | GPIOBIT(PA3) | \
	GPIOBIT(PA4) | GPIOBIT(PA5) | GPIOBIT(PA6) | GPIOBIT(PA17) | \
	GPIOBIT(PA18) | GPIOBIT(PA22) | GPIOBIT(PA23))

// enable default state of outputs that need to start high
#define DEF_PIOA_SODR	0

// enable default state of outputs that need to start low
#define DEF_PIOA_CODR	GPIOBIT(LCD_BKL)

// enable all pins that start as outputs
#define DEF_PIOA_OER	GPIOBIT(LCD_BKL)

// PORT B						   BIT	A/B	DIR	SENSE	DESCRIPTION
#define VCC5_EN 	GPIO(32,0)	// PB0  -	Out Lo	0=Enable 5V, 1=Disable 5V
#define USB_SCON	GPIO(33,1)	// PB1 - 	In	Hi	USB Soft Connect
#define USR_LED 	GPIO(34,2)	// PB2  -	Out Lo	User LED
#define B_SDD1	 	GPIO(35,3)	// PB3  B	-	-	SD Card Data 1 
#define B_SDD2	 	GPIO(36,4)	// PB4  B	-	-	SD Card Data 2
#define B_SDD3	 	GPIO(37,5)	// PB5  B	-	-	SD Card Data 3
#define PB6	 		GPIO(38,6)	// PB6  - 	-	-	Expansion GPIO/TFI/TIOA3
#define IDE_RST	 	GPIO(39,7)	// PB7  - 	In	Lo	IDE Reset 
#define IDE_IRQ	 	GPIO(40,8)	// PB8  - 	In	Hi	IDE Interrupt	 
#define PB9	 		GPIO(41,9)	// PB9  - 	-	-	Expansion GPIO/RD1/TIOB4
#define PB10	 	GPIO(42,10)	// PB10 - 	-	-	Expansion GPIO/RK1/TIOA5
#define PB11	 	GPIO(43,11)	// PB11 - 	-	-	Expansion GPIO/RF1/TIOB5
#define B_ETXD2	 	GPIO(44,12)	// PB12 B	-	-	Ethernet Transmit Data 2
#define B_ETXD3	 	GPIO(45,13)	// PB13 B	-	-	Ethernet Transmit Data 3
#define B_ETXER	 	GPIO(46,14)	// PB14 B	-	-	Ethernet Transmit Error
#define B_ERXD2	 	GPIO(47,15)	// PB15 B	-	-	Ethernet Receive Data  
#define B_ERXD3	 	GPIO(48,16)	// PB16 B	-	-	Ethernet Receive Data 
#define B_ERXDV	 	GPIO(49,17)	// PB17 B	-	-	Ethernet Receive Data Valid
#define B_ECOL	 	GPIO(50,18)	// PB18 B	-	-	Ethernet Collision
#define B_ERXCK	 	GPIO(51,19)	// PB19 B	-	-	Ethernet Receive Clock
#define PB20	 	GPIO(52,20)	// PB20 -	-	-	Expansion GPIO/TXD1
#define PB21	 	GPIO(53,21)	// PB21 -	-	-	Expansion GPIO/RXD1
#define PB22	 	GPIO(54,22)	// PB22 -	-	-	Expansion GPIO/SCK1
#define LCD_GPIO 	GPIO(55,23)	// PB23 -	-	-		 
#define PB24	 	GPIO(56,24)	// PB24 -	-	-	Expansion GPIO/USART1 CTS
#define IDE_PWR	 	GPIO(57,25)	// PB25 -	Out	Hi	IDE Power Enable, 1 = On	
#define PB26	 	GPIO(58,26)	// PB26 -	-	-	Expansion GPIO/USART1 RTS
#define A_PCK0		GPIO(59,27)	// PB27 A	-	-	Periph Clock 0 - Bus Clock
#define USB_CD		GPIO(60,28)	// PB28 -	In	Lo	USB Device Cable Detect
#define PB29		GPIO(61,29)	// PB29 -	In	Hi	Expansion GPIO/IRQ0
// PB30 - Does not exist							
// PB31 - Does not exist							

// default settings for PORT B
// Peripheral A select - USART 1, Peripheral Clock 0
#define DEF_PIOB_ASR	(GPIOBIT(A_PCK0))

// Peripheral B select - SD Card 0 and Ethernet
#define DEF_PIOB_BSR	\
	(GPIOBIT(B_SDD1) | GPIOBIT(B_SDD2) | GPIOBIT(B_SDD3) | \
	GPIOBIT(B_ETXD2) | GPIOBIT(B_ETXD3) | GPIOBIT(B_ETXER) | \
	GPIOBIT(B_ERXD2) | GPIOBIT(B_ERXD3) | GPIOBIT(B_ERXDV) | \
	GPIOBIT(B_ECOL) | GPIOBIT(B_ERXCK))

 // enable pullups where desired - usually just expansion GPIO
#define DEF_PIOB_PUER	\
	(GPIOBIT(PB6) | GPIOBIT(PB9) | GPIOBIT(PB10) | GPIOBIT(PB11) | \
	GPIOBIT(PB20) | GPIOBIT(PB21) | GPIOBIT(PB22) | GPIOBIT(PB24) | \
	GPIOBIT(PB26) | GPIOBIT(PB29))

// enable default state of outputs that need to start high
#define DEF_PIOB_SODR	(GPIOBIT(VCC5_EN) | GPIOBIT(USR_LED))

// enable default state of outputs that need to start low
#define DEF_PIOB_CODR	0

// enable all pins that start as outputs
#define DEF_PIOB_OER	\
	(GPIOBIT(VCC5_EN) | GPIOBIT(USR_LED) | GPIOBIT(IDE_PWR))

// PORT C						   BIT	A/B	DIR	SENSE	DESCRIPTION
#define PHY_INT		GPIO(64,0)	// PC0	-	In	Lo	LT971A PHY Interrupt
#define PIRQ 		GPIO(65,1)	// PC1	-	In	Lo	Touch Interrupt
#define SD_PWR 		GPIO(66,2)	// PC2	-	Out	Hi	SD Card Power On
#define CF_CD 		GPIO(67,3)	// PC3	-	In	Lo	Compact Flash Detect
#define SD_CD		GPIO(68,4)	// PC4	-	In	Lo	SD Card Detect
#define SD_WP		GPIO(69,5)	// PC5	-	In	Lo	SD Card Write Protect
#define A_WAIT	 	GPIO(70,6)	// PC6	A	-	-	WAIT Input	
#define PC7		 	GPIO(71,7)	// PC7	-	-	-	Unavailable
#define PC8	 		GPIO(72,8)	// PC8  - 	In	-	Unavailable
#define CF_RST 		GPIO(73,9)	// PC9	-	Out	Lo	Compact Flash Reset
#define CF_RW	 	GPIO(74,10)	// PC10 - 	-	-	Compact Flash R/W
#define CF_CE1 		GPIO(75,11)	// PC11 A 	-	-	Compact Flash Card Enable 1
#define CF_CE2	 	GPIO(76,12)	// PC12 A 	-	-	Compact Flash Card Enable 2
#define PC13	 	GPIO(77,13)	// PC13	-	-	-	Expansion GPIO/CS7
#define CF_PWR 		GPIO(78,14)	// PC14	-	Out	Hi	Compact Flash Power Enable
#define PC15	 	GPIO(79,15)	// PC15	-	-	-	Unavailable
#define A_D16	 	GPIO(80,16)	// PC16	A	-	-	Data Bus 16
#define A_D17	 	GPIO(81,17)	// PC17	A	-	-	Data Bus 17
#define A_D18	 	GPIO(82,18)	// PC18	A	-	-	Data Bus 18
#define A_D19	 	GPIO(83,19)	// PC19	A	-	-	Data Bus 19
#define A_D20	 	GPIO(84,20)	// PC20	A	-	-	Data Bus 20
#define A_D21	 	GPIO(85,21)	// PC21	A	-	-	Data Bus 21
#define A_D22	 	GPIO(86,22)	// PC22	A	-	-	Data Bus 22
#define A_D23	 	GPIO(87,23)	// PC23	A	-	-	Data Bus 23
#define A_D24	 	GPIO(88,24)	// PC24	A	-	-	Data Bus 24
#define A_D25	 	GPIO(89,25)	// PC25	A	-	-	Data Bus 25
#define A_D26	 	GPIO(90,26)	// PC26	A	-	-	Data Bus 26
#define A_D27	 	GPIO(91,27)	// PC27	A	-	-	Data Bus 27
#define A_D28	 	GPIO(92,28)	// PC28	A	-	-	Data Bus 28
#define A_D29	 	GPIO(93,29)	// PC29	A	-	-	Data Bus 29
#define A_D30	 	GPIO(94,30)	// PC30	A	-	-	Data Bus 30
#define A_D31	 	GPIO(95,31)	// PC31	A	-	-	Data Bus 31

// default settings for PORT C
// Peripheral A select - Wait, A23 and Databus
#define DEF_PIOC_ASR	\
	(GPIOBIT(A_WAIT) | GPIOBIT(CF_CE1) | GPIOBIT(CF_CE2) | 0xffff0000)

// Peripheral B select - None
#define DEF_PIOC_BSR	0

// enable pullups where desired - usually just expansion GPIO
#define DEF_PIOC_PUER	\
	(GPIOBIT(PC7) | GPIOBIT(PC8) | GPIOBIT(PC13) | GPIOBIT(PC15))

// enable default state of outputs that need to start high
#define DEF_PIOC_SODR	0

// enable default state of outputs that need to start low
#define DEF_PIOC_CODR	(GPIOBIT(CF_PWR) | GPIOBIT(SD_PWR))

// enable all pins that start as outputs
#define DEF_PIOC_OER	(GPIOBIT(CF_RST) | GPIOBIT(CF_PWR) | GPIOBIT(SD_PWR))

// PORT D					   		   BIT	A/B	DIR	SENSE	DESCRIPTION
#define PD0	 		GPIO(96,0)		// PD0	-	-	-	Unavailable
#define PD1	 		GPIO(97,1)		// PD1	-	-	-	Unavailable
#define PD2	 		GPIO(98,2)		// PD2	-	-	-	Unavailable
#define PD3	 		GPIO(99,3)		// PD3	-	-	-	Unavailable
#define PD4	 		GPIO(100,4)		// PD4	-	-	-	Unavailable
#define ONE_WIRE	GPIO(101,5)		// PD5	-	In	Lo	DS2401 1-Wire Device
#define PD6			GPIO(102,6)		// PD6	-	-	-	Unavailable
#define B_ETMSYNC	GPIO(103,7)		// PD7	B	-	-	ETM Sync 
#define B_ETMCLK	GPIO(104,8)		// PD8	B	-	-	ETM Clock
#define B_ETMPS0	GPIO(105,9)		// PD9	B	-	-	ETM Proc Status Bit 0
#define B_ETMPS1	GPIO(106,10)	// PD10	B	-	-	ETM Proc Status Bit 1
#define B_ETMPS2	GPIO(107,11)	// PD11	B	-	-	ETM Proc Status Bit 2
#define B_ETMPK0	GPIO(108,12)	// PD12	B	-	-	ETM Packet Bit 0
#define B_ETMPK1	GPIO(109,13)	// PD13	B	-	-	ETM Packet Bit 1
#define B_ETMPK2	GPIO(110,14)	// PD14	B	-	-	ETM Packet Bit 2
#define B_ETMPK3	GPIO(111,15)	// PD15	B	-	-	ETM Packet Bit 3
#define B_ETMPK4	GPIO(112,16)	// PD16	B	-	-	ETM Packet Bit 4
#define B_ETMPK5	GPIO(113,17)	// PD17	B	-	-	ETM Packet Bit 5
#define B_ETMPK6	GPIO(114,18)	// PD18	B	-	-	ETM Packet Bit 6
#define B_ETMPK7	GPIO(115,19)	// PD19	B	-	-	ETM Packet Bit 7
#define B_ETMPK8	GPIO(116,20)	// PD20	B	-	-	ETM Packet Bit 8
#define B_ETMPK9	GPIO(117,21)	// PD21	B	-	-	ETM Packet Bit 9
#define B_ETMPK10	GPIO(118,22)	// PD22	B	-	-	ETM Packet Bit 10
#define B_ETMPK11	GPIO(119,23)	// PD23	B	-	-	ETM Packet Bit 11
#define B_ETMPK12	GPIO(120,24)	// PD24	B	-	-	ETM Packet Bit 12
#define B_ETMPK13	GPIO(121,25)	// PD25	B	-	-	ETM Packet Bit 13
#define B_ETMPK14	GPIO(122,26)	// PD26	B	-	-	ETM Packet Bit 14
#define B_ETMPK15	GPIO(123,27)	// PD27	B	-	-	ETM Packet Bit 15
// PD28 Does Not Exist
// PD29 Does Not Exist
// PD30 Does Not Exist
// PD31 Does Not Exist

// default settings for PORT D
// Peripheral A select - none
#define DEF_PIOD_ASR	0

// Peripheral B select - Embedded Trace Module
// We probably shouldn't enable ETM by default as it can be turned on from JTAG.
#define DEF_PIOD_BSR	0

 // enable pullups on all but the two CD Card status bits
#define DEF_PIOD_PUER   0

// enable default state of outputs that need to start high
#define DEF_PIOD_SODR	0

// enable default state of outputs that need to start low
#define DEF_PIOD_CODR	0

// enable all pins that start as outputs
#define DEF_PIOD_OER	0


#endif // #ifndef CSB637_GPIO_H


