//==========================================================================
//
// csb737_gpio.h
//
// Author(s):
//		Michael Kelly, Cogent Computer Systems, Inc.
//		Luis Torrico, Cogent Computer Systems, Inc.
// Modifications: 
//		10/15/2005 Updated to conform to uMon's generic GPIO structure
//		12/27/2007 Modified from CSB637 settings to support CSB737
// Description:
//		This file contains the board level to vitual GPIO bit assignments
//
// The AT91SAM9263 GPIO's are spread across five 32-bit ports A-E.
// For bits that have an alternate function assigned, we use the bit 
// position within the respective port as we need that to setup the
// registers.  But for bits that are used as GPIOs we assign a 
// virtual GPIO number which is used by the GPIO functions
//
// This section also defines the bit setting for the alternate
// functions of the pins.  If we want alternate function A, we prepend
// an A, for function B, a B.  This insures we will get an error, or
// at least it will be very visible if we try to assign a function
// to the wrong side.  Note the A or B is not related to the Port A
// B, just the alternate function.
//
// Finally we build the default values to be written into the GPIO registers
// in gpio_init (in csb737_gpio.c).

#include "umongpio.h"

#ifndef CSB737_GPIO_H
#define CSB737_GPIO_H

// PORT A				BIT					A/B	DIR	SENSE	DESCRIPTION
#define B_SPI0_MISO		GPIO(0,0)	// PA0  B	In	-		SPI0 Master In - Slave Out
#define	B_SPI0_MOSI		GPIO(1,1)	// PA1  B	Out	-		SPI0 Master Out - Slave In
#define B_SPI0_CLK		GPIO(2,2)	// PA2  B	Out	Lo		SPI0 Clock
#define GPIO6			GPIO(3,3)	// PA3  -	I/O	-		GPIO or SPI0 Chip Select 1 (Peripheral B)
#define SPI0_RDY		GPIO(4,4)	// PA4  -	In	-		SPI0 Ready or GPIO
#define B_SPI0_CS0		GPIO(5,5)	// PA5  B	Out	-		SPI0 Chip Select 0
#define A_SD_CLK		GPIO(6,6)	// PA6  A 	Out	-		SD Card Clock
#define A_SD_CMD		GPIO(7,7)	// PA7  A	Out	-		SD Card Command
#define A_SD_D0			GPIO(8,8)	// PA8  A	I/O	-		SD Card Data 0
#define A_SD_D1			GPIO(9,9)	// PA9  A	I/O	-		SD Card Data 1
#define A_SD_D2			GPIO(10,10)	// PA10 A	I/O	-		SD Card Data 2
#define A_SD_D3			GPIO(11,11)	// PA11 A	I/O	-		SD Card Data 3
#define PA12			GPIO(12,12)	// PA12 -	In	-		Unconnected
#define A_CAN0_TXD		GPIO(13,13)	// PA13 A	Out	-		CAN Transmit Data
#define A_CAN0_RXD		GPIO(14,14)	// PA14 A	In	-		CAN Receive Data
#define GPIO1			GPIO(15,15)	// PA15 -	I/O	-		GPIO or IRQ1 (Peripheral B)
#define PA16			GPIO(16,16)	// PA16 -	In	-		Unconnected
#define PA17			GPIO(17,17)	// PA17 - 	In	-		Unconnected
#define PA18			GPIO(18,18)	// PA18 - 	In	-		Unconnected
#define PA19			GPIO(19,19)	// PA19 - 	In	-		Unconnected
#define PA20			GPIO(20,20)	// PA20 - 	In	-		Unconnected
#define PA21			GPIO(21,21)	// PA21 - 	In	-		Unconnected
#define NRDY			GPIO(22,22)	// PA22 - 	In	Hi		NAND Ready/Busy (1 = Ready)
#define PA23			GPIO(23,23)	// PA23 - 	In	-		Unconnected
#define PA24			GPIO(24,24)	// PA24 - 	In	-		Unconnected
#define PA25			GPIO(25,25)	// PA25 - 	In	-		Unconnected
#define A_U0_TXD		GPIO(26,26)	// PA26 A 	Out	-		UART0 Transmit Data
#define A_U0_RXD		GPIO(27,27)	// PA27 A 	In 	- 		UART0 Receive Data 
#define A_U0_RTS		GPIO(28,28)	// PA28 A 	Out - 		UART0 Request to Send
#define A_U0_CTS		GPIO(29,29)	// PA29 A 	In 	- 		UART0 Clear to SendTransmit Data
#define PA30			GPIO(30,30)	// PA30 - 	In	- 		Unconnected
#define PA31			GPIO(31,31)	// PA31 - 	In	-		Unconnected

// default settings for PORT A
// Peripheral A select - SD Card 0, UART 0 and CAN 0
#define DEF_PIOA_ASR	\
	(GPIOBIT(A_SD_CLK) | GPIOBIT(A_SD_CMD) | GPIOBIT(A_SD_D0) | \
	GPIOBIT(A_SD_D1) | GPIOBIT(A_SD_D2) | GPIOBIT(A_SD_D3) | \
	GPIOBIT(A_CAN0_TXD) | GPIOBIT(A_CAN0_RXD) | GPIOBIT(A_U0_TXD) | \
	GPIOBIT(A_U0_RXD) | GPIOBIT(A_U0_RTS) | GPIOBIT(A_U0_CTS))

// Peripheral B select - SPI 0
#define DEF_PIOA_BSR	\
	(GPIOBIT(B_SPI0_MISO) | GPIOBIT(B_SPI0_MOSI) | GPIOBIT(B_SPI0_CLK) | \
	GPIOBIT(B_SPI0_CS0))

// Enable pullups where desired - usually just unused GPIO and all inputs
#define DEF_PIOA_PUER	\
	(GPIOBIT(SPI0_RDY) | GPIOBIT(A_CAN0_RXD) | GPIOBIT(NRDY) | GPIOBIT(A_U0_RXD) | \
	GPIOBIT(A_U0_CTS) | GPIOBIT(PA12) | GPIOBIT(PA16) | GPIOBIT(PA17) | GPIOBIT(PA18) | \
	GPIOBIT(PA19) | GPIOBIT(PA20) | GPIOBIT(PA21) | GPIOBIT(PA23) | GPIOBIT(PA24) | \
	GPIOBIT(PA25) | GPIOBIT(PA30) | GPIOBIT(PA31))

// Enable default state of GPIO outputs that need to start high
#define DEF_PIOA_SODR	0

// Enable default state of GPIO outputs that need to start low
#define DEF_PIOA_CODR	0

// Enable all GPIO pins that start as outputs
#define DEF_PIOA_OER	0

// PORT B								BIT	A/B	DIR	SENSE	DESCRIPTION
#define A_AC_SYNC 		GPIO(32,0)	// PB0  A	In	-		AC97 SYNC from Codec
#define A_AC_BCLK		GPIO(33,1)	// PB1	A	In	-		AC97 Bit Clock from Codec
#define A_AC_SDOUT		GPIO(34,2)	// PB2  A	Out -		AC97 Serial Data to Codec
#define A_AC_SDIN	 	GPIO(35,3)	// PB3  A	In	-		AC97 Serial Data from Codec
#define A_I2C_SDA	 	GPIO(36,4)	// PB4  A	I/O	-		Two-Wire Interface Data
#define A_I2C_SCL	 	GPIO(37,5)	// PB5  A	I/O	-		Two-Wire Interface Clock
#define A_SSI_FRM	 	GPIO(38,6)	// PB6  A 	I/O	-		SSI Frame
#define A_SSI_CLK	 	GPIO(39,7)	// PB7  A 	I/O	-		SSI Clock 
#define A_SSI_TXD	 	GPIO(40,8)	// PB8  A 	Out	-		SSI Transmit	 
#define A_SSI_RXD	 	GPIO(41,9)	// PB9  A 	In	-		SSI Receive
#define A_SSI_MCLK	 	GPIO(42,10)	// PB10 A 	I/O	-		SSI Master Clock
#define SPI1_RDY	 	GPIO(43,11)	// PB11 - 	In	-		SSI Ready or GPIO
#define A_SPI1_MISO	 	GPIO(44,12)	// PB12 A	In	-		SPI1 Master In - Slave Out
#define A_SPI1_MOSI	 	GPIO(45,13)	// PB13 A	Out	-		SPI1 Master Out - Slave In
#define A_SPI1_CLK	 	GPIO(46,14)	// PB14 A	Out	-		SPI1 Clock
#define A_SPI1_CS0	 	GPIO(47,15)	// PB15 A	Out	Lo		SPI1 Chip Select 0
#define GPIO7	 		GPIO(48,16)	// PB16 -	I/O	-		GPIO or SPI1 Chip Select 1 (Peripheral A)
#define GPIO2	 		GPIO(49,17)	// PB17 -	I/O	-		GPIO or Timer A IO2 (Peripheral B)
#define GPIO3	 		GPIO(50,18)	// PB18 -	I/O	-		GPIO or Timer B IO2 (Peripheral B)
#define AC_RST	 		GPIO(51,19)	// PB19 -	Out	Lo		AC97 Codec Reset (0 = Reset)
#define PIRQ	 		GPIO(52,20)	// PB20 -	In	Lo		Touch (PEN) Interrupt, Low True
#define EX_IRQ	 		GPIO(53,21)	// PB21 -	In	Lo		Expansion Interrupt, Low True
#define CF_RDY	 		GPIO(54,22)	// PB22 -	In	-		Compact Flash Card Ready or Interrupt
#define CF_CD 			GPIO(55,23)	// PB23 -	In	Lo		Compact Flash Card Detect (0 = Card In)	 
#define CF_RST	 		GPIO(56,24)	// PB24 -	Out	Hi		Compact Flash Card Reset (1 = Reset)
#define PB25	 		GPIO(57,25)	// PB25 -	In	-		Unconnected
#define PB26	 		GPIO(58,26)	// PB26 -	In	-		Unconnected
#define GPIO8			GPIO(59,27)	// PB27 -	I/O	-		GPIO or PWM2 (Peripheral B)
#define GPIO9			GPIO(60,28)	// PB28 -	I/O	-		GPIO or Timer 0 Clock (Peripheral B)
#define LCD_BKL			GPIO(61,29)	// PB29 -	Out	Hi		LCD Backlight, 1 = on
#define I2C_INT			GPIO(62,30)	// PB30	-	In	Lo		Shared I2C Devices Interrupt, Low True	
#define LOW_PWR			GPIO(63,31)	// PB31	-	Out	Lo		Drive 0 to indicate sleep to ram mode	

// default settings for PORT B
// Peripheral A select - AC97, SSI and SPI 1
#define DEF_PIOB_ASR	\
	(GPIOBIT(A_AC_SYNC) | GPIOBIT(A_AC_BCLK) | GPIOBIT(A_AC_SDOUT) | \
	GPIOBIT(A_AC_SDIN) | GPIOBIT(A_I2C_SDA) | GPIOBIT(A_I2C_SCL) | \
	GPIOBIT(A_SSI_FRM) | GPIOBIT(A_SSI_CLK) | GPIOBIT(A_SSI_TXD) | \
	GPIOBIT(A_SSI_RXD) | GPIOBIT(A_SSI_MCLK) | GPIOBIT(A_SPI1_MISO) | \
	GPIOBIT(A_SPI1_MOSI) | GPIOBIT(A_SPI1_CLK) | GPIOBIT(A_SPI1_CS0))

// Peripheral B select - None
#define DEF_PIOB_BSR	0

// Enable pullups where desired - usually just unused GPIO and all inputs
#define DEF_PIOB_PUER	\
	(GPIOBIT(A_AC_SYNC) | GPIOBIT(A_AC_BCLK) | GPIOBIT(A_AC_SDIN) | \
	GPIOBIT(A_SSI_RXD) | GPIOBIT(SPI1_RDY) | GPIOBIT(A_SPI1_MISO) | \
	GPIOBIT(PIRQ) | GPIOBIT(EX_IRQ) | GPIOBIT(CF_RDY) | GPIOBIT(CF_CD) | \
	GPIOBIT(PB25) | GPIOBIT(PB26) | GPIOBIT(I2C_INT))

// Enable default state of GPIO outputs that need to start high
#define DEF_PIOB_SODR	(GPIOBIT(AC_RST) | GPIOBIT(LOW_PWR))

// Enable default state of GPIO outputs that need to start low
#define DEF_PIOB_CODR	(GPIOBIT(CF_RST) | GPIOBIT(LCD_BKL))
	
// Enable all GPIO pins that start as outputs
#define DEF_PIOB_OER	\
	(GPIOBIT(AC_RST) | GPIOBIT(CF_RST) | GPIOBIT(LCD_BKL) | GPIOBIT(LOW_PWR))

// Enable Multi Drive pins
#define DEF_PIOB_MDER	(GPIOBIT(A_I2C_SDA) | GPIOBIT(A_I2C_SCL))

// PORT C							   BIT	A/B	DIR	SENSE	DESCRIPTION
#define A_LCD_VS		GPIO(64,0)	// PC0	A	Out	-		LCD Vertical Sync
#define A_LCD_HS		GPIO(65,1)	// PC1	A	Out	-		LCD Horizontal Sync
#define A_LCD_PCLK		GPIO(66,2)	// PC2	A	Out	-		LCD Pixal Clock
#define A_LCD_OE		GPIO(67,3)	// PC3	A	Out	-		LCD Output Enable
#define B_LCD_B1		GPIO(68,4)	// PC4	B	Out	-		LCD Blue Bit 1 (LCDD3)
#define B_LCD_B2		GPIO(69,5)	// PC5	B	Out	-		LCD Blue Bit 2 (LCDD4)
#define B_LCD_B3 		GPIO(70,6)	// PC6	B	Out	-		LCD Blue Bit 3 (LCDD5)
#define B_LCD_B4 		GPIO(71,7)	// PC7	B	Out	-		LCD Blue Bit 4 (LCDD6)
#define B_LCD_B5		GPIO(72,8)	// PC8  B	Out	-		LCD Blue Bit 5 (LCDD7)
#define B_LCD_G0		GPIO(73,9)	// PC9	B	Out	-		LCD Green Bit 0 (LCDD10)
#define B_LCD_G1 		GPIO(74,10)	// PC10 B	Out	-		LCD Green Bit 1 (LCDD11)
#define B_LCD_G2		GPIO(75,11)	// PC11 B 	Out	-		LCD Green Bit 2 (LCDD12)
#define B_LCD_G3 		GPIO(76,12)	// PC12 B 	Out	-		LCD Green Bit 3 (LCDD13)
#define B_LCD_G4 		GPIO(77,13)	// PC13	B	Out	-		LCD Green Bit 4 (LCDD14)
#define B_LCD_G5		GPIO(78,14)	// PC14	B	Out	-		LCD Green Bit 5 (LCDD15)
#define B_LCD_R1 		GPIO(79,15)	// PC15	B	Out	-		LCD Red Bit 1 (LCDD19)
#define B_LCD_R2 		GPIO(80,16)	// PC16	B	Out	-		LCD Red Bit 2 (LCDD20)
#define B_LCD_R3 		GPIO(81,17)	// PC17	B	Out	-		LCD Red Bit 3 (LCDD21)
#define B_LCD_R4 		GPIO(82,18)	// PC18	B	Out	-		LCD Red Bit 4 (LCDD22)
#define B_LCD_R5 		GPIO(83,19)	// PC19	-	In	-		LCD Red Bit 5 (LCDD23)
#define PC20	 		GPIO(84,20)	// PC20	-	In	-		Unconnected
#define PC21	 		GPIO(85,21)	// PC21	-	In	-		Unconnected
#define PC22	 		GPIO(86,22)	// PC22	-	In	-		Unconnected
#define PC23	 		GPIO(87,23)	// PC23	-	In	-		Unconnected
#define PC24	 		GPIO(88,24)	// PC24	-	In	-		Unconnected
#define B_E_CRSDV 		GPIO(89,25)	// PC25	B	Out	-		Carrier Sense/Data Valid in RMII Mode
#define PC26	 		GPIO(90,26)	// PC26	-	In	-		Unconnected
#define PC27	 		GPIO(91,27)	// PC27	-	In	-		Unconnected
#define GPIO4	 		GPIO(92,28)	// PC28	-	I/O	-		GPIO OR PWM0 (Peripheral A)
#define GPIO5	 		GPIO(93,29)	// PC29	-	I/O	-		GPIO OR PWM2 (Peripheral B)
#define A_D_RXD	 		GPIO(94,30)	// PC30	A	In	-		Debug UART Receive (via RS232)
#define A_D_TXD	 		GPIO(95,31)	// PC31	A	Out	-		Debug UART Transmit (via RS232)

// default settings for PORT C
// Peripheral A select - LCD and Debug UART
#define DEF_PIOC_ASR	\
	(GPIOBIT(A_LCD_VS) | GPIOBIT(A_LCD_HS) | GPIOBIT(A_LCD_PCLK) | GPIOBIT(A_LCD_OE) | \
	GPIOBIT(A_D_RXD) | GPIOBIT(A_D_TXD))
	

// Peripheral B select - LCD Data and Ethernet CRSDV
#define DEF_PIOC_BSR	\
	(GPIOBIT(B_LCD_B1) | GPIOBIT(B_LCD_B2) | GPIOBIT(B_LCD_B3) | GPIOBIT(B_LCD_B4) | \
	GPIOBIT(B_LCD_B5) | GPIOBIT(B_LCD_G0) | GPIOBIT(B_LCD_G1) | GPIOBIT(B_LCD_G2) | \
	GPIOBIT(B_LCD_G3) | GPIOBIT(B_LCD_G4) | GPIOBIT(B_LCD_G5) | GPIOBIT(B_LCD_R1) | \
	GPIOBIT(B_LCD_R2) | GPIOBIT(B_LCD_R3) | GPIOBIT(B_LCD_R4) | GPIOBIT(B_LCD_R5) | \
	GPIOBIT(B_E_CRSDV))

// Enable pullups where desired - usually just unused GPIO and all inputs
#define DEF_PIOC_PUER	\
	(GPIOBIT(PC20) | GPIOBIT(PC21) | GPIOBIT(PC22) | GPIOBIT(PC23) | \
	GPIOBIT(PC24) | GPIOBIT(PC26) | GPIOBIT(PC27) | GPIOBIT(A_D_RXD))

// Enable default state of GPIO outputs that need to start high
#define DEF_PIOC_SODR	0

// Enable default state of GPIO outputs that need to start low
#define DEF_PIOC_CODR	0

// Enable all GPIO pins that start as outputs
#define DEF_PIOC_OER	0

// PORT D					   			   BIT	A/B	DIR	SENSE	DESCRIPTION
#define A_U1_TXD	 	GPIO(96,0)		// PD0	A	Out	-		UART1 Transmit Data
#define A_U1_RXD	 	GPIO(97,1)		// PD1	A	In	-		UART1 Receive Data
#define U1_RTS		 	GPIO(98,2)		// PD2	-	Out	-		UART1 Request To Send (GPIO Mode) or UART2 Transmit Data (Peripheral A)
#define U1_CTS		 	GPIO(99,3)		// PD3	-	In	-		UART1 Clear To Send (GPIO Mode) or UART2 Receive Data (Peripheral A)
#define GPIO0			GPIO(100,4)		// PD4	-	In	-		GPIO or FIQ (Peripheral A) User LED on CSB702
#define A_WAIT			GPIO(101,5)		// PD5	A	In	Lo		Expansion and Compact Flash Wait
#define A_EBI0_CS4		GPIO(102,6)		// PD6	A	Out	Lo		Set *CS4 in Compact Flash Mode
#define PD7				GPIO(103,7)		// PD7	-	In	-		Unconnected
#define A_CF_CE1		GPIO(104,8)		// PD8	A	Out Lo		Compact Flash Chip Enable 1
#define A_CF_CE2		GPIO(105,9)		// PD9	A	Out	Lo		Compact Flash Chip Enable 2
#define PD10			GPIO(106,10)	// PD10	-	In	-		Unconnected
#define A_EBI0_CS2		GPIO(107,11)	// PD11	A	Out	Lo		SODIMM Expansion Chip Select
#define A_EBI0_A23		GPIO(108,12)	// PD12	A	Out	-		SAM9263 Address Bit 23
#define A_EBI0_A24		GPIO(109,13)	// PD13	A	Out	-		SAM9263 Address Bit 24
#define A_EBI0_A25		GPIO(110,14)	// PD14	A	Out	-		SAM9263 Address Bit 25
#define A_EBI0_CS3		GPIO(111,15)	// PD15	A	Out	-		NAND Flash Chip Select
#define A_EBI0_D16		GPIO(112,16)	// PD16	A	I/O	-		EBI0 Data Bits 16
#define A_EBI0_D17		GPIO(113,17)	// PD17	A	I/O	-		EBI0 Data Bits 17
#define A_EBI0_D18		GPIO(114,18)	// PD18	A	I/O	-		EBI0 Data Bits 18
#define A_EBI0_D19		GPIO(115,19)	// PD19	A	I/O	-		EBI0 Data Bits 19
#define A_EBI0_D21		GPIO(116,20)	// PD20	A	I/O	-		EBI0 Data Bits 20
#define A_EBI0_D22		GPIO(117,21)	// PD21	A	I/O	-		EBI0 Data Bits 21
#define A_EBI0_D23		GPIO(118,22)	// PD22	A	I/O	-		EBI0 Data Bits 22
#define A_EBI0_D24		GPIO(119,23)	// PD23	A	I/O	-		EBI0 Data Bits 23
#define A_EBI0_D25		GPIO(120,24)	// PD24	A	I/O	-		EBI0 Data Bits 24
#define A_EBI0_D26		GPIO(121,25)	// PD25	A	I/O	-		EBI0 Data Bits 25
#define A_EBI0_D27		GPIO(122,26)	// PD26	A	I/O	-		EBI0 Data Bits 26
#define A_EBI0_D28		GPIO(123,27)	// PD27	A	I/O	-		EBI0 Data Bits 27
#define A_EBI0_D29		GPIO(124,28)	// PD28	A	I/O	-		EBI0 Data Bits 28
#define A_EBI0_D20		GPIO(125,29)	// PD29	A	I/O	-		EBI0 Data Bits 29
#define A_EBI0_D30		GPIO(126,30)	// PD30	A	I/O	-		EBI0 Data Bits 30
#define A_EBI0_D31		GPIO(127,31)	// PD31	A	I/O	-		EBI0 Data Bits 31

// default settings for PORT D
// Peripheral A select - UART1, CF Wait, CF CE's, EBI0 CS's, EBI0 A23-25 and EBI0 Data 16-31
#define DEF_PIOD_ASR	\
	(GPIOBIT(A_U1_TXD) | GPIOBIT(A_U1_RXD) | GPIOBIT(A_WAIT) | GPIOBIT(A_EBI0_CS4) | \
	GPIOBIT(A_CF_CE1) | GPIOBIT(A_CF_CE2) | GPIOBIT(A_EBI0_CS2) | GPIOBIT(A_EBI0_A23) | \
	GPIOBIT(A_EBI0_A24) | GPIOBIT(A_EBI0_A25) | GPIOBIT(A_EBI0_CS3) | 0xffff0000)

// Peripheral B select - Embedded Trace Module
// We probably shouldn't enable ETM by default as it can be turned on from JTAG.
#define DEF_PIOD_BSR	0

// Enable pullups where desired - usually just unused GPIO and all inputs
#define DEF_PIOD_PUER   \
	(GPIOBIT(A_U1_RXD) | GPIOBIT(U1_CTS) | GPIOBIT(GPIO0) | \
	GPIOBIT(A_WAIT) | GPIOBIT(PD7) | GPIOBIT(PD10))

// Enable default state of GPIO outputs that need to start high
#define DEF_PIOD_SODR	(GPIOBIT(U1_RTS))

// Enable default state of GPIO outputs that need to start low
#define DEF_PIOD_CODR	0

// Enable all GPIO pins that start as outputs
#define DEF_PIOD_OER	(GPIOBIT(U1_RTS))

// PORT E					   			   BIT	A/B	DIR	SENSE	DESCRIPTION
#define A_VIP_D0	 	GPIO(128,0)		// PE0	A	In	-		Video Input Port Data Bit 0
#define A_VIP_D1	 	GPIO(129,1)		// PE1	A	In	-		Video Input Port Data Bit 1
#define A_VIP_D2	 	GPIO(130,2)		// PE2	A	In	-		Video Input Port Data Bit 2
#define A_VIP_D3	 	GPIO(131,3)		// PE3	A	In	-		Video Input Port Data Bit 3
#define A_VIP_D4		GPIO(132,4)		// PE4	A	In	-		Video Input Port Data Bit 4
#define A_VIP_D5		GPIO(133,5)		// PE5	A	In	-		Video Input Port Data Bit 5
#define A_VIP_D6		GPIO(134,6)		// PE6	A	In	-		Video Input Port Data Bit 6
#define A_VIP_D7		GPIO(135,7)		// PE7	A	In	-		Video Input Port Data Bit 7
#define A_VIP_PCLK		GPIO(136,8)		// PE8	A	In	-		Video Input Port Pixel Clock
#define A_VIP_HS		GPIO(137,9)		// PE9	A	In	-		Video Input Port Horizontal Sync
#define A_VIP_VS		GPIO(138,10)	// PE10	A	In	-		Video Input Port Vertical Sync
#define A_VIP_MCLK		GPIO(139,11)	// PE11	A	In	-		Video Input Port Master Clock
#define B_VIP_D8		GPIO(140,12)	// PE12	B	In	-		Video Input Port Data Bit 8
#define B_VIP_D9		GPIO(141,13)	// PE13	B	In	-		Video Input Port Data Bit 9
#define PE14			GPIO(142,14)	// PE14	-	In	-		Unconnected
#define PE15			GPIO(143,15)	// PE15	-	In	-		Unconnected
#define PE16			GPIO(144,16)	// PE16	-	In	-		Unconnected
#define PE17			GPIO(145,17)	// PE17	-	In	-		Unconnected
#define PE18			GPIO(146,18)	// PE18	-	In	-		Unconnected
#define PE19			GPIO(147,19)	// PE19	-	In	-		Unconnected
#define PE20			GPIO(148,20)	// PE20	-	In	-		Unconnected
#define A_E_REFCLK		GPIO(149,21)	// PE21	A	In	-		Ethernet Reference Clock (50Mhz), only in RMII Mode
#define PE22			GPIO(150,22)	// PE22	-	In	-		Unconnected
#define A_E_TXD0		GPIO(151,23)	// PE23	A	Out	-		RMII Ethernet Transmit Data Bit 0
#define A_E_TXD1		GPIO(152,24)	// PE24	A	Out	-		RMII Ethernet Transmit Data Bit 1
#define A_E_RXD0		GPIO(153,25)	// PE25	A	In	-		RMII Ethernet Receive Data Bit 0
#define A_E_RXD1		GPIO(154,26)	// PE26	A	In	-		RMII Ethernet Receive Data Bit 1
#define A_E_RXER		GPIO(155,27)	// PE27	A	In	-		RMII Ethernet Receive Error
#define A_E_TXEN		GPIO(156,28)	// PE28	A	Out	-		RMII Ethernet Transmit Enable
#define A_E_MDC			GPIO(157,29)	// PE29	A	Out	-		RMII Ethernet Management Bus Clock
#define A_E_MDIO		GPIO(158,30)	// PE30	A	I/O	-		RMII Ethernet Management Bus Data
#define PE31			GPIO(159,31)	// PE31	-	In	-		Unconnected

// default settings for PORT E
// Peripheral A select - Video Input Port and RMII Ethernet
#define DEF_PIOE_ASR	\
	(GPIOBIT(A_VIP_D0) | GPIOBIT(A_VIP_D1) | GPIOBIT(A_VIP_D2) | GPIOBIT(A_VIP_D3) | \
	GPIOBIT(A_VIP_D4) | GPIOBIT(A_VIP_D5) | GPIOBIT(A_VIP_D6) | GPIOBIT(A_VIP_D7) | \
	GPIOBIT(A_VIP_PCLK) | GPIOBIT(A_VIP_HS) | GPIOBIT(A_VIP_VS) | GPIOBIT(A_VIP_MCLK) | \
	GPIOBIT(A_E_REFCLK) | GPIOBIT(A_E_TXD0) | GPIOBIT(A_E_TXD1) | GPIOBIT(A_E_RXD0) | \
	GPIOBIT(A_E_RXD1) | GPIOBIT(A_E_RXER) | GPIOBIT(A_E_TXEN) | GPIOBIT(A_E_MDC) | GPIOBIT(A_E_MDIO))

// Peripheral B select - Video Input Port
#define DEF_PIOE_BSR	(GPIOBIT(B_VIP_D8) | GPIOBIT(B_VIP_D9))

// Enable pullups where desired - usually just unused GPIO and all inputs
#define DEF_PIOE_PUER   \
	(GPIOBIT(A_VIP_D0) | GPIOBIT(A_VIP_D1) | GPIOBIT(A_VIP_D2) | GPIOBIT(A_VIP_D3) | \
	GPIOBIT(A_VIP_D4) | GPIOBIT(A_VIP_D5) | GPIOBIT(A_VIP_D6) | GPIOBIT(A_VIP_D7) | \
	GPIOBIT(A_VIP_PCLK) | GPIOBIT(A_VIP_HS) | GPIOBIT(A_VIP_VS) | GPIOBIT(A_VIP_MCLK) | \
	GPIOBIT(B_VIP_D8) | GPIOBIT(B_VIP_D9) | GPIOBIT(PE14) | GPIOBIT(PE15) | \
	GPIOBIT(PE16) | GPIOBIT(PE17) | GPIOBIT(PE18) | GPIOBIT(PE19) | \
	GPIOBIT(PE20) | GPIOBIT(A_E_REFCLK) | GPIOBIT(PE22) | GPIOBIT(A_E_RXD0) | \
	GPIOBIT(A_E_RXD1) | GPIOBIT(A_E_RXER) | GPIOBIT(PE31))

// Enable default state of GPIO outputs that need to start high
#define DEF_PIOE_SODR	0

// Enable default state of GPIO outputs that need to start low
#define DEF_PIOE_CODR	0

// Enable all GPIO pins that start as outputs
#define DEF_PIOE_OER	0


#endif // #ifndef CSB737_GPIO_H


