/* fec.c:
 * This code is specific to the MPC852T
 * As pages are mentioned throughout this text, it is referencing the
 * MPC866UM.pdf (MPC866 PowerQUICK Family Refernce Manual.
 *
 * General notice:
 * This code is part of a boot-monitor package developed as a generic base
 * platform for embedded system designs.  As such, it is likely to be
 * distributed to various projects beyond the control of the original
 * author.  Please notify the author of any enhancements made or bugs found
 * so that all may benefit from the changes.  In addition, notification back
 * to the author will allow the new user to pick up changes that may have
 * been made by other users after this version of the code was distributed.
 * 
 * Author:	Ed Sutter
 * email:	esutter@lucent.com
 * phone:	908-582-2351
 */
#include "config.h"
#if INCLUDE_ETHERNET & USE_FEC
#include "cpuio.h"
#include "ether.h"
#include "stddefs.h"
#include "genlib.h"
#include "arch_ppc.h"
#include "mpc860.h"
#include "masks860.h"
#include "timer.h"

#define USE_INTERNAL_DESCRIPTORS

/* Rcv ctrl:
 */
#define LOOPBACK	1
#define DRT			2
#define MII_MODE	4
#define PROMISCUOUS	8
#define BC_REJ		16

/* Xmt ctrl:
 */
#define FULL_DUPLEX 4
#define HBC			2
#define GTS			1

/* Ethernet ctrl:
 */
#define EN_RESET	1
#define EN_ENABLE	2
#define EN_PINMUX	4
#define DES_ACTIVE 0x01000000

/* definitions taken from table 44-7:
 */
#define FEC_ADDR_LOW		0xe00
#define FEC_ADDR_HIGH		0xe04
#define FEC_HASH_TABLE_HIGH	0xe08
#define FEC_HASH_TABLE_LOW	0xe0c
#define FEC_R_DES_START		0xe10
#define FEC_X_DES_START		0xe14
#define FEC_R_BUFF_SIZE		0xe18
#define FEC_ECNTRL			0xe40
#define FEC_IEVENT			0xe44
#define FEC_IMASK			0xe48
#define FEC_IVEC			0xe4c
#define FEC_R_DES_ACTIVE	0xe50
#define FEC_X_DES_ACTIVE	0xe54
#define FEC_MII_DATA		0xe80
#define FEC_MII_SPEED		0xe84
#define FEC_R_BOUND			0xecc
#define FEC_R_FSTART		0xed0
#define FEC_X_WMRK			0xee4
#define FEC_X_FSTART		0xeec
#define FEC_FUN_CODE		0xf34
#define FEC_R_CNTRL			0xf44
#define FEC_R_HASH			0xf48
#define FEC_X_CNTRL			0xf84

/* Event bits:
 */
#define HBERR	0x80000000
#define BABR	0x40000000
#define BABT	0x20000000
#define GRA		0x10000000
#define TFINT	0x08000000
#define TXB		0x04000000
#define RFINT	0x02000000
#define RXB		0x01000000
#define MII		0x00800000
#define EBERR	0x00400000

/* Buffer descriptor error bits for FEC:
 */
#define BD_EMPTY		0x8000
#define BD_TRUNCATED	0x0001
#define BD_OVERRUN		0x0002
#define BD_CRCERR		0x0004
#define BD_SHORT		0x0008
#define BD_NONOCTET		0x0010
#define BD_LARGE		0x0020

extern	EPPC *IMMRBase;

static uchar xbuf[XBUFCNT][1600];	/* Transmit & receive buffers pointed */
static uchar rbuf[RBUFCNT][1600];	/* to by the buffer descriptors. */

struct	enet_bd *tbptr, *rbptr, *prevtbptr;
struct enet_bd e_bdtbl[XBUFCNT+RBUFCNT];

/* FEC-specific error counters:
 */
int EtherHeartbeatError;
int EtherBabblingRcvError;
int EtherGracefulStopCompletions;
int EtherXmtFrameCompletions;
int EtherXmtBufferUpdates;
int EtherRcvBufferUpdates;
int EtherMIIInterruptCount;
int EtherBusError;
int	EtherNONOCTETCnt;
int	EtherSHORTCnt;
int	EtherTOOBIGCnt;
int	EtherTRNCCnt;
int	EtherCRCERRCnt;
int EtherOVRNCnt;

void
showFECDescriptors(void)
{
	int i;
	struct	enet_bd	*tbd, *rbd;

	i = 0;
	tbd = (struct enet_bd *)(IMMRBase->fec_or_rsvd.fec.x_des_start);
	printf("TBDbase: 0x%lx\n",(ulong)tbd);
	while(1) {
		printf("TBD_%d:\n",i++);
		printf(" status=0x%x, length=0x%x, baddr= 0x%lx\n",
			tbd->status,tbd->length,(ulong)tbd->baddr);
		if (tbd->status & BD_WRAP)
			break;
		tbd++;
	}

	i = 0;
	rbd = (struct enet_bd *)(IMMRBase->fec_or_rsvd.fec.r_des_start);
	printf("RBDbase: 0x%lx\n",(ulong)rbd);
	while(1) {
		printf("RBD_%d:\n",i++);
		printf(" status=0x%x, length=0x%x, baddr= 0x%lx\n",
			rbd->status,rbd->length,(ulong)rbd->baddr);
		if (rbd->status & BD_WRAP)
			break;
		rbd++;
	}
}

/* ShowEtherdevStats():
 * Called by the command "ether stat" on the monitor's command line.
 * This function and a common ShowEthernetStats() are used to dump
 * generic as well as device-specific ethernet driver statistics.
*/
void
ShowEtherdevStats()
{
	printf("FEC-based ethernet interface\n");
	if (IMMRBase->fec_or_rsvd.fec.r_cntrl & LOOPBACK) {
		printf("Receive mode: ");
		if (IMMRBase->fec_or_rsvd.fec.r_cntrl & LOOPBACK)
			printf("loopback ");
		if (IMMRBase->fec_or_rsvd.fec.r_cntrl & DRT)
			printf("disable rcv-on-xmt ");
		if (IMMRBase->fec_or_rsvd.fec.r_cntrl & MII_MODE)
			printf("mii_mode ");
		if (IMMRBase->fec_or_rsvd.fec.r_cntrl & PROMISCUOUS)
			printf("promiscuous ");
		if (IMMRBase->fec_or_rsvd.fec.r_cntrl & BC_REJ)
			printf("bcast_reject ");
		putchar('\n');
	}

	printf("EtherHeartbeatErrors:         %d\n",EtherHeartbeatError);
	printf("EtherBabblingRcvErrors:       %d\n",EtherBabblingRcvError);
	printf("EtherGracefulStopCompletions: %d\n",EtherGracefulStopCompletions);
	printf("EtherXmtFrameCompletions:     %d\n",EtherXmtFrameCompletions);
	printf("EtherXmtBufferUpdates:        %d\n",EtherXmtBufferUpdates);
	printf("EtherRcvBufferUpdates:        %d\n",EtherRcvBufferUpdates);
	printf("EtherMIIInterruptCounts:      %d\n",EtherMIIInterruptCount);
	printf("EtherBusErrors:               %d\n",EtherBusError);
	printf("EtherNONOCTETCnt:             %d\n",EtherNONOCTETCnt);
	printf("EtherSHORTCnt:                %d\n",EtherSHORTCnt);
	printf("EtherTOOBIGCnt:               %d\n",EtherTOOBIGCnt);
	printf("EtherTRNCCnt:                 %d\n",EtherTRNCCnt);
	printf("EtherCRCERRCnt:               %d\n",EtherCRCERRCnt);
	printf("EtherOVERRUNCnt:              %d\n",EtherOVRNCnt);

	if (EtherVerbose)
		showFECDescriptors();
}

/* DisableEtherdev():
 * Called by DisableEthernet() to do the device-specific portion of 
 * the turn-down.
 */
void
DisableEtherdev()
{
	IMMRBase->fec_or_rsvd.fec.ecntrl &= ~EN_ENABLE;
	EtherIsActive = 0;
	IPMonCmdActive = 0;
}

void
enreset(void)
{
	struct elapsed_tmr tmr;

	startElapsedTimer(&tmr,5000);
	/* Set RESET bit, then wait for it to clear:
	 */
	IMMRBase->fec_or_rsvd.fec.ecntrl |= EN_RESET;
	while(IMMRBase->fec_or_rsvd.fec.ecntrl & EN_RESET) {
		if (msecElapsed(&tmr)) {
			printf("enreset() timeout\n");
			break;
		}
	}
}

int
eninit(void)
{
	volatile int	i;
	volatile ulong mac;
	volatile uchar *cp;
	volatile struct	enet_bd	*tbd, *rbd;

	/* Sanity check of the data structure overlaying the FEC:
	 */
	if (((long)(&IMMRBase->fec_or_rsvd.fec.addr_low) & 0xfff) != FEC_ADDR_LOW)
		printf("FEC structure error: addrlow\n");
	if (((long)(&IMMRBase->fec_or_rsvd.fec.mii_data) & 0xfff) != FEC_MII_DATA)
		printf("FEC structure error: mii_data\n");
	if (((long)(&IMMRBase->fec_or_rsvd.fec.ecntrl) & 0xfff) != FEC_ECNTRL)
		printf("FEC structure error: ecntrl\n");
	if (((long)(&IMMRBase->fec_or_rsvd.fec.r_bound) & 0xfff) != FEC_R_BOUND)
		printf("FEC structure error: r_bound\n");
	if (((long)(&IMMRBase->fec_or_rsvd.fec.x_cntrl) & 0xfff) != FEC_X_CNTRL)
		printf("FEC structure error: x_cntrl\n");

	/* Properly configure the muxed pins associated with the MII...
	 * This information is in appendix H for the MPC852T.
	 * PD3-PD15 must be configured for MII mode.
	 * Clear PD[UT] to disable Utopia and enable MII.
	 * The IRQ7/MII_TX_CLK is put in MII mode when the EN_ENABLE bit
	 * is set in ecntrl register.
	 */
	IMMRBase->pio_pdpar &=  ~0xc000;	/* Clear UTOPIA & ATM bits */
	IMMRBase->pio_pdpar |=  0x1fff;		/* Set to non-pio usage */
	IMMRBase->pio_pddir |=  0x1fff;		/* MII mode */
	IMMRBase->pio_pcpar |=  0x0008;
	IMMRBase->pio_pcdir |=  0x0008;
	IMMRBase->siu_simask &=	~0x0002;	/* Clear IRQ7 SIMask bit */

	/* This code follows the suggested steps outlined in section 44.4.2
	 * of the MPC866 User manual...
	 */

	/* Disable interrupts:
	 */
	IMMRBase->fec_or_rsvd.fec.imask = 0x0;

	/* Clear all events (1 clears an event).
	 */
	IMMRBase->fec_or_rsvd.fec.ievent = 0xffffffff;
	
	IMMRBase->fec_or_rsvd.fec.ivec = 0;
	
	cp = (uchar *)&mac;
	cp[1] = BinEnetAddr[5];
	cp[0] = BinEnetAddr[4];
	cp[2] = 0;
	cp[3] = 0;
	IMMRBase->fec_or_rsvd.fec.addr_high = mac;
	cp[3] = BinEnetAddr[3];
	cp[2] = BinEnetAddr[2];
	cp[1] = BinEnetAddr[1];
	cp[0] = BinEnetAddr[0];
	IMMRBase->fec_or_rsvd.fec.addr_low = mac;

	IMMRBase->fec_or_rsvd.fec.hash_table_high = 0x0;
	IMMRBase->fec_or_rsvd.fec.hash_table_low = 0x0;

	/* Initialize Buffer Descriptors (located in DPRAM):
	 */
#ifdef USE_INTERNAL_DESCRIPTORS
	tbd = tbptr = IMMRBase->qcp_or_ud.umd.e_bdtbl;
	rbd = rbptr = &IMMRBase->qcp_or_ud.umd.e_bdtbl[XBUFCNT];
#else
	tbd = tbptr = e_bdtbl;
	rbd = rbptr = &e_bdtbl[XBUFCNT];
#endif

	IMMRBase->fec_or_rsvd.fec.r_des_start = (ulong)rbptr;
	IMMRBase->fec_or_rsvd.fec.x_des_start = (ulong)tbptr;
	IMMRBase->fec_or_rsvd.fec.r_buff_size = 1520;

	IMMRBase->fec_or_rsvd.fec.r_cntrl = MII_MODE;
//	IMMRBase->fec_or_rsvd.fec.r_cntrl = (MII_MODE | PROMISCUOUS | LOOPBACK);
	IMMRBase->fec_or_rsvd.fec.x_cntrl = FULL_DUPLEX;
	IMMRBase->fec_or_rsvd.fec.fun_code = 0x78000000;

	/* NOTE: the documentation on this mii_speed setting seems inconsistent.
	   The value written here was given to me by Anatoli.
	 */
	IMMRBase->fec_or_rsvd.fec.mii_speed = 0x003e;
	IMMRBase->fec_or_rsvd.fec.ecntrl = (EN_ENABLE | EN_PINMUX);

	/* Initialize transmit buffer ring:
	 */
	for (i=0;i<XBUFCNT;i++,tbd++) {
		ulong tmp;

		tbd->status = 0;
		tbd->length = 0;
		tbd->baddr = xbuf[i];
		tmp = (ulong)tbd->baddr;
		tmp |= 0x3f;
		tbd->baddr = (uchar *)tmp;
		tbd->baddr += 1;
		if (i == (XBUFCNT-1))
			tbd->status |= BD_WRAP;	/* Set WRAP bit on final BD */
	}

	/* Initialize receive buffer ring:
	 */
	for(i=0;i<RBUFCNT;i++,rbd++) {
		ulong tmp;

		rbd->status = BD_READY;
		rbd->length = 0;
		rbd->baddr = rbuf[i];
		tmp = (ulong)rbd->baddr;
		tmp |= 0x3f;
		rbd->baddr = (uchar *)tmp;
		rbd->baddr += 1;
		if (i == (RBUFCNT-1))
			rbd->status |= BD_WRAP;	/* Set WRAP bit on final BD */
	}

	IMMRBase->fec_or_rsvd.fec.r_des_active = DES_ACTIVE;
	return(0);
}

/* EtherdevStartup():
 * Called by EthernetStartup().  Initialize all device-specific 
 * counters, reset and initialized the ethernet device.
*/
int
EtherdevStartup(int verbose)
{
	EtherTRNCCnt = 0;
	EtherOVRNCnt = 0;
	EtherBusError = 0;
	EtherSHORTCnt = 0;
	EtherTOOBIGCnt = 0;
	EtherCRCERRCnt = 0;
	EtherNONOCTETCnt = 0;
	EtherHeartbeatError = 0;
	EtherBabblingRcvError = 0;
	EtherXmtBufferUpdates = 0;
	EtherRcvBufferUpdates = 0;
	EtherMIIInterruptCount = 0;
	EtherXmtFrameCompletions = 0;
	EtherGracefulStopCompletions = 0;

	/* Put ethernet controller in reset: */
	enreset();

	monDelay(1000);

	/* Initialize controller: */
	eninit();

	return(0);
}

/* enselftest():
 * 1. Enable ethernet interface.
 * 2. Put the PHY in loopback.
 * 3. Send a packet, and see if that packet is received.
 * 4. Disable the ethernet interface.
 * Return 1 if passed, else -1.
 */
int
enselftest(int verbose)
{
	int passed = 1;

	IMMRBase->fec_or_rsvd.fec.r_cntrl |= LOOPBACK;
	return(passed);
}

/* polletherdev():
 * Called by pollethernet() to do the device-specific portion of the
 * ethernet polling.
 */
int
polletherdev(void)
{
	int ptot;
	vulong	event;
	struct	enet_bd *tmprbp;

	/* Check for incoming events from the FEC.
	 */
	if ((event = IMMRBase->fec_or_rsvd.fec.ievent) != 0) {
		IMMRBase->fec_or_rsvd.fec.ievent = event;
		if (event & HBERR)
			EtherHeartbeatError++;
		if (event & BABR)
			EtherBabblingRcvError++;
		if (event & BABT)
			EtherGracefulStopCompletions++;
		if (event & TFINT)
			EtherXmtFrameCompletions++;
		if (event & TXB)
			EtherXmtBufferUpdates++;
		if (event & RXB)
			EtherRcvBufferUpdates++;
		if (event & MII)
			EtherMIIInterruptCount++;
		if (event & EBERR)
			EtherBusError++;
	}
	
	ptot = 0;
	if (!(rbptr->status & BD_EMPTY)) {
		/* Copy rbptr to local variable so that polletherdev() can
		 * be called by processPACKET() without worrying reusing the
		 * same buffer descriptor pointer...
		 */
		tmprbp = rbptr;

		/* Increment the buffer descriptor pointer...  If WRAP, then set
		 * it back to the top; otherwise, increment to next buffer
		 * descriptor entry.
		 */
		if (rbptr->status & BD_WRAP)
			rbptr = (struct enet_bd *)(IMMRBase->fec_or_rsvd.fec.r_des_start);
		else 
			rbptr++;

		processPACKET((struct ether_header *)tmprbp->baddr,tmprbp->length);
		tmprbp->status |= BD_EMPTY;
		IMMRBase->fec_or_rsvd.fec.r_des_active = DES_ACTIVE;
		EtherRFRAMECnt++;
		ptot++;
	}
	return(ptot);
}

/* getXmitBuffer():
 * Called by various points in the IP/UDP/ICMP code for a buffer that
 * will ultimately be the next output packet.
 */
uchar *
getXmitBuffer()
{
	struct elapsed_tmr tmr;

	/* Wait for the next buffer descriptor to become available... */
	startElapsedTimer(&tmr,5000);
	while(tbptr->status & BD_READY) {
		if (msecElapsed(&tmr)) {
			printf("getXmitBuffer() timeout\n");
			break;
		}
	}
	return(tbptr->baddr);
}

/* sendBuffer():
 * Set the flag within the current transmit buffer descriptor indicating
 * that it is ready to be transmitted.  Then increment the BD pointer to 
 * the next buffer descriptor.
 */
int
sendBuffer(int length)
{
	if (EtherVerbose &  SHOW_OUTGOING)
		printPkt((struct ether_header *)tbptr->baddr,length,ETHER_OUTGOING);

	/* Keep stats: */
	if (EtherXFRAMECnt) {
		if (prevtbptr->status & BD_OVERRUN) {
			EtherOVRNCnt++;
		}
		else {
			if (prevtbptr->status & BD_NONOCTET)
				EtherNONOCTETCnt++;
			if (prevtbptr->status & BD_SHORT)
				EtherSHORTCnt++;
			if (prevtbptr->status & BD_LARGE)
				EtherTOOBIGCnt++;
			if (prevtbptr->status & BD_TRUNCATED)
				EtherTRNCCnt++;
			if ((prevtbptr->status & (BD_LAST | BD_CRCERR)) ==
				(BD_LAST | BD_CRCERR))
				EtherCRCERRCnt++;
		}
	}

	EtherXFRAMECnt++;

	/* Make the current transmit buffer ready and populate the length.
	 * Tell the FEC to check the output ring buffer.
	 */
	tbptr->length = (ushort)length;
	tbptr->status |= (BD_READY | BD_LAST | BD_TC);
	IMMRBase->fec_or_rsvd.fec.x_des_active = DES_ACTIVE;

	/* Keep track of this BD so that the status can be looked at when the */
	/* next transmission is done. */
	prevtbptr = tbptr;

	/* If WRAP, then back to top; else, just increment to next BD. */
	if (tbptr->status & BD_WRAP)
		tbptr = (struct enet_bd *)(IMMRBase->fec_or_rsvd.fec.x_des_start);
	else
		tbptr++;

	return(0);
}

/* Disable/Enable either broadcast or promiscuous mode... */

void
disableBroadcastReception()
{
	IMMRBase->fec_or_rsvd.fec.r_cntrl |= BC_REJ;
}

void
enableBroadcastReception()
{
	IMMRBase->fec_or_rsvd.fec.r_cntrl &= ~BC_REJ;
}

void
enablePromiscuousReception()
{
	IMMRBase->fec_or_rsvd.fec.r_cntrl |= PROMISCUOUS;
}

void
disablePromiscuousReception()
{
	IMMRBase->fec_or_rsvd.fec.r_cntrl &= ~PROMISCUOUS;
}

char *
extGetEtherAdd()
{
	return((char *)0);
}

char *
extGetIpAdd()
{
	return((char *)0);
}

#endif
