/* cpuio.c:
 *	CPU/Board Specific IO.
 *
 *	This file contains the IO functions required by Micro Monitor
 *	that are unique to each CPU/Board combination.
 *
 *	General notice:
 *	This code is part of a boot-monitor package developed as a generic base
 *	platform for embedded system designs.  As such, it is likely to be
 *	distributed to various projects beyond the control of the original
 *	author.  Please notify the author of any enhancements made or bugs found
 *	so that all may benefit from the changes.  In addition, notification back
 *	to the author will allow the new user to pick up changes that may have
 *	been made by other users after this version of the code was distributed.
 *
 *	Author:	Torsten Tetz, Bosch Security Systems
 *	        Mark Jonas, Bosch Security Systems, mark.jonas@de.bosch.com
 *	        Joerg Tuermer, Bosch Security Systems, joerg.tuermer@de.bosch.com
 *	based on template/cpuio.c file by Ed Sutter
 */
#include "config.h"
#include "genlib.h"
#include "cpuio.h"
#include "timerdev.h"
#include "uart.h"

/* ASM macros */
#ifndef STRINGIFY
   #define STRINGIFY(x) #x
#endif

#define LDC(variable, creg)            \
   asm volatile                        \
   (                                   \
      "ldc %[val],"STRINGIFY(creg)     \
      : /* no outputs */               \
      : [val] "r" (variable)           \
   )

#define STC(creg, variable)            \
   asm volatile                        \
   (                                   \
      "stc "STRINGIFY(creg)",%[val]"   \
      : [val] "=r" (variable)          \
      : /* no inputs */                \
   )

/* setup_mpr2():
 * Initialize the board.
 * Set chip selects, port mulitplexing.
 */
void 
setup_mpr2(void)
{
#if (PLATFORM_MP_R2==2)
	/* Other function for port:   G0        G1        G2        G3        G4       G5    G6    P4
	 * Set Pin Select Register A: /PCC_CD1, /PCC_CD2, PCC_BVD1, PCC_BVD2, /IOIS16, IRQ4, IRQ5, USB1d_SUSPEND
	 */
	ctrl_outw(0xAABC, PSELA);  /*       10        10        10        10       10    11    11             00 */

	/* Other function for port:   T3          T4          V3         V4         E5       E6       
	 * Set Pin Select Register B: /SCIF0_RTS, /SCIF0_CTS, LCD_VCPWC, LCD_VEPWC, IIC_SDA, IIC_SCL, Reserved
	 */
	ctrl_outw(0x3C00, PSELB);  /*          0           0         11         11        0        0  00000000 */

	/* Other function for port:   U0         U1         V1         V2
	 * Set Pin Select Register C: SIOF1_SCK, SIOF1_RxD, SCIF1_RxD, SCIF1_TxD, Reserved
	 */
	ctrl_outw(0x0000, PSELC);  /*        00         00         00         00  00000000 */

	/* Other function for port:             U2                   U3                    U4                    V0
	 * Set Pin Select Register D: Reserved, SIOF1_TxD, Reserved, SIOF1_MCLK, Reserved, SIOF1_SYNC, Reserved, SCIF1_SCK, Reserved
	 */
	ctrl_outw(0x0000, PSELD);  /*        0         00        00          00        00          00        00         00         0 */

	/* Set USB Tranceiver Control: Reserved, DRV, Reserved, USB_TRANS, USB_SEL
	 */
	ctrl_outw(0x0101, UTRCTL);  /*  0000000    1    000000          0        1 */
 	
	/* Set USB Clock Control:        USSCS,   USSTB,   Reserved   (HighByte always A5) 
	 */
	ctrl_outw(0xA5C0, UCLKCR_W);   /*  110        0        0000 */


	/* CS2: Ethernet (0x08000000 - 0x0bffffff) */
	ctrl_outl(0x36db0400, CS2BCR);		/* 4 idle cycles, normal space, 16 bit data bus */
	ctrl_outl(0x000003c0, CS2WCR);		/* (WR:8), no ext. wait	*/

	/* CS4: CAN1 (0xb0000000 - 0xb3ffffff) */
	ctrl_outl(0x00000200, CS4BCR);		/* no idle cycles, normal space, 8 bit data bus */
	ctrl_outl(0x00100981, CS4WCR);		/* (SW:1.5 WR:3 HW:1.5), ext. wait		BAS:1  1.5cycles ReadAccess:3cycles  1.5cycles */

	/* CS5a: CAN2 (0xb4000000 - 0xb5ffffff) */
	ctrl_outl(0x00000200, CS5ABCR);		/* no idle cycles, normal space, 8 bit data bus */
	ctrl_outl(0x00000981, CS5AWCR);		/* (SW:1.5 WR:3 HW:1.5), ext. wait */

	/* CS5b: CAN3 (0xb6000000 - 0xb7ffffff) */
	ctrl_outl(0x00000200, CS5BBCR);		/* no idle cycles, normal space, 8 bit data bus */
	ctrl_outl(0x00100981, CS5BWCR);		/* (SW:1.5 WR:3 HW:1.5), ext. wait */

	/* CS6a: Rotary (0xb8000000 - 0xb9ffffff) */
	ctrl_outl(0x00000200, CS6ABCR);		/* no idle cycles, normal space, 8 bit data bus */
	ctrl_outl(0x000009C1, CS6AWCR);		/* (SW:1.5 WR:3 HW:1.5), no ext. wait */

	/* IRQ0..3 are independent interrupt pins
	 * IRQ5: BUSY - touch screen - (falling edge sensitive)
	 * IRQ4: PENIRQ - touch screen - (rising edge sensitive)
	 * IRQ3: LAN_IRQ - Ethernet - (low level sensitive)
	 * IRQ2: CAN3_IRQ - CAN - (low level sensitive)
	 * IRQ1: CAN2_IRQ - CAN - (low level sensitive)
	 * IRQ0: CAN1_IRQ - CAN - (low level sensitive)
	 */
	ctrl_outw(ICR1_IRQ5_FALL | ICR1_IRQ4_RISE | ICR1_IRQ3_LOW | ICR1_IRQ2_LOW | ICR1_IRQ1_LOW | ICR1_IRQ0_LOW, ICR1);

	/* PINT4, PINT6 and PINT7 are used as independent interrupt pins
	 * PINT7: /PWR2 (low level sensitive)
	 * PINT6: /PWR1 (low level sensitive)
	 * PINT4: FLASH_READY (high level sensitive)
	 */
	ctrl_bit_clear_outw( (ICR2_PINT7S | ICR2_PINT6S), ICR2);			/* Set polarity of PINT interrupts to low level */
	ctrl_bit_set_outw( ICR2_PINT4S, ICR2);					/* Set polarity of PINT interrupts to high level */
	ctrl_outw( PINTER_PINT7E | PINTER_PINT6E | PINTER_PINT4E, PINTER);	/* Enable PINT interrupts */

	ctrl_outb(0x30, PMDR);	/* CURRENT_SINK=off, PWR_SWITCH=off, LAN_SPEED=100MBit, LAN_RESET=off, BUZZER=off, LCD_BL=off */
	                      	/*                                   1                  1                                     */

	/* PMDR: LAN_RESET=on */
	ctrl_bit_clear_outb(0x10, PMDR);

	/* PMDR: LAN_RESET=off */
	ctrl_bit_set_outb(0x10, PMDR);

	/* set peripheral multiplexer for MP-R2 A.2
	 * OTH:  (00) Other fuction
	 * GPO:  (01) General Purpose Output
	 * GPI:  (11) General Purpose Input
	 * GPI+: (10) General Purpose Input with internal pull-up
	 *-------------------------------------------------------
	 * A7 GPO(LED8);   A6 GPO(LED7);   A5 GPO(LED6);   A4 GPO(LED5);
	 * A3 GPO(LED4);   A2 GPO(LED3);   A1 GPO(LED2);   A0 GPO(LED1);
	 */
	ctrl_outw(0x5555, PACR);	/* 01 01 01 01 01 01 01 01 */

	/* B7 GPO(RST4);   B6 GPO(RST3);   B5 GPO(RST2);   B4 GPO(RST1);
	 * B3 GPO(PB3);    B2 GPO(PB2);    B1 GPO(PB1);    B0 GPO(PB0);
	 */
	ctrl_outw(0x5555, PBCR);	/* 01 01 01 01 01 01 01 01 */

	/* C7 GPO(PC7);    C6 GPO(PC6);    C5 GPO(PC5);    C4 GPO(PC4);
	 * C3 LCD_DATA3;   C2 LCD_DATA2;   C1 LCD_DATA1;   C0 LCD_DATA0;
	 */
	ctrl_outw(0x5500, PCCR);	/* 01 01 01 01 00 00 00 00 */

	/* D7 GPO(PD7);   D6 GPO(PD6);   D5 GPO(PD5);   D4 GPO(PD4);
	 * D3 GPO(PD3);   D2 GPO(PD2);   D1 GPO(PD1);   D0 GPO(PD0);
	 */
	ctrl_outw(0x5555, PDCR);	/* 01 01 01 01 01 01 01 01 */

	/* E7 (x);       E6 GPI(nu);   E5 GPI(nu);   E4 LCD_M_DISP;
	 * E3 LCD_CL1;   E2 LCD_CL2;   E1 LCD_DON;   E0 LCD_FLM;
	 */
	ctrl_outw(0x2800, PECR);	/* 00 10 10 00 00 00 00 00 */

	/* F7 (x);           F6 DA1(VLCD);       F5 DA0(nc);      F4 AN3;
	 * F3 AN2(MID_AD);   F2 AN1(EARTH_AD);   F1 AN0(TEMP);    F0 GPI+(nc);
	 */
	ctrl_outw(0x0002, PFCR);	/* 00 00 00 00 00 00 00 10 */

	/* G7 (x);         G6 IRQ5(TOUCH_BUSY);   G5 IRQ4(TOUCH_IRQ);   G4 GPI(KEY2);
	 * G3 GPI(KEY1);   G2 GPO(LED11);         G1 GPO(LED10);        G0 GPO(LED9);
	 */
	ctrl_outw(0x03D5, PGCR);	/* 00 00 00 11 11 01 01 01 */

	/* H7 (x);              H6 /RAS(BRAS);        H5 /CAS(BCAS);   H4 CKE(BCKE);
	 * H3 GPO(EARTH_OFF);   H2 GPO(EARTH_TEST);   H1 USB2_PWR;     H0 USB1_PWR;
	 */
	ctrl_outw(0x0050, PHCR);	/* 00 00 00 00 01 01 00 00 */

	/* J7 (x);       J6 AUDCK;     J5 ASEBRKAK;   J4 AUDATA3;
	 * J3 AUDATA2;   J2 AUDATA1;   J1 AUDATA0;    J0 AUDSYNC;
	 */
	ctrl_outw(0x0000, PJCR);	/* 00 00 00 00 00 00 00 00 */

	/* K7 (x);            K6 (x);            K5 (x);         K4 (x);
	 * K3 PINT7(/PWR2);   K2 PINT6(/PWR1);   K1 PINT5(nu);   K0 PINT4(FLASH_READY);
	 */
	ctrl_outw(0x00FB, PKCR);	/* 00 00 00 00 11 11 10 11 */

	/* L7 TRST;    L6 TMS;     L5 TDO;      L4 TDI;
	 * L3 TCK;     L2 (x);     L1 (x);      L0 (x);
	 */
	ctrl_outw(0x0000, PLCR);	/* 00 00 00 00 00 00 00 00 */

	/* M7 GPO(CURRENT_SINK);   M6 GPO(PWR_SWITCH);   M5 GPO(LAN_SPEED);   M4 GPO(LAN_RESET);
	 * M3 GPO(BUZZER);         M2 GPO(LCD_BL);       M1 CS5B(CAN3_CS);    M0 GPI+(nc);
	 */
	ctrl_outw(0x5552, PMCR);	/* 01 01 01 01 01 01 00 10 */

	/* P7 (x);             P6 (x);              P5 (x);              P4 GPO(nu);
	 * P3 IRQ3(LAN_IRQ);   P2 IRQ2(CAN3_IRQ);   P1 IRQ1(CAN2_IRQ);   P0 IRQ0(CAN1_IRQ);
	 */
	ctrl_outw(0x0100, PPCR);	/* 00 00 00 01 00 00 00 00 */
	ctrl_outb(0x10, PPDR);		/* No current flow through pullup */

	/* R7 A25;   R6 A24;   R5 A23;   R4 A22;
	 * R3 A21;   R2 A20;   R1 A19;   R0 A0;
	 */
	ctrl_outw(0x0000, PRCR);	/* 00 00 00 00 00 00 00 00 */

	/* S7 (x);               S6 (x);         S5 (x);         S4 GPO(EEPROM_CS2);
	 * S3 GPO(EEPROM_CS1);   S2 SIOF0_TXD;   S1 SIOF0_RXD;   S0 SIOF0_SCK;
	 */
	ctrl_outw(0x0140, PSCR);	/* 00 00 00 01 01 00 00 00 */

	/* T7 (x);             T6 (x);        T5 (x);        T4 GPO(COM1_CTS);
	 * T3 GPO(COM1_RTS);   T2 COM1_TXD;   T1 COM1_RXD;   T0 GPO(WDOG)
	 */
	ctrl_outw(0x0141, PTCR);	/* 00 00 00 01 01 00 00 01 */

	/* U7 (x);             U6 (x);         U5 (x);         U4 GPI+(/AC_FAULT);
	 * U3 GPO(TOUCH_CS);   U2 TOUCH_TXD;   U1 TOUCH_RXD;   U0 TOUCH_SCK;
	 */
	ctrl_outw(0x0340, PUCR);	/* 00 00 00 11 01 00 00 00 */

	/* V7 (x);         V6 (x);        V5 (x);        V4 GPO(MID2);
	 * V3 GPO(MID1);   V2 CARD_TxD;   V1 CARD_RxD;   V0 GPI+(/BAT_FAULT);
	 */
	ctrl_outw(0x0143, PVCR);	/* 00 00 00 01 01 00 00 11 */

#else
	#error Unknown revision of PLATFORM_MP_R2
#endif
}

/* devInit():
 * The common version of start.c calls devInit(baud) here to process 
 * all device initializations.
 */
int
devInit(int baud)
{
#if INCLUDE_HWTMR
	/* Start the "Timer Unit"(TMU chanel 0) for the use of the board
	 * based timer inside common uMon code. Since it is needed to 
	 * initialize the UART, it must be started first.
	 */
	init_timer0();
#endif

	/* Initialize the UART so we can print messages. */
	initUART(baud);
	return(0);
}

/* initCPUio():
 * If CPU IO wasn't initialized in reset.S, do it here...
 * This provides a "C-level" IO init opportunity. 
 */
void
initCPUio(void)
{
	/* Initialize the interrupt control registers: */
	*IPRA = 0;
	*IPRB = 0;
	*IPRC = 0;
	*IPRD = 0;
	*IPRE = 0;
	*IPRF = 0;
	*IPRG = 0;
	*IPRH = 0;
	*IPRI = 0;
	*IPRJ = 0;
	*ICR0 = 0;
	*ICR1 = 0;
	*ICR2 = 0;
	setup_mpr2();
}

/* target_reset():
 * This function leads to coldstart(), to be found in "reset.S",
 * causing a complete restart of uMon.
 */
void
target_reset(void)
{
	coldstart();
}

/* putvbr():
 * Set the vector base register.
 * This is needed for setting up the exception handling.
 * Used in "except_sh3.c".
 */
void
putvbr(unsigned long value)
{
	LDC(value,vbr);
}

/* putsr():
 * Set the status register.
 * This is needed for setting up the exception handling.
 * Used in "except_sh3.c".
 */
void
putsr(unsigned long value)
{
	LDC(value,sr);
}

#if 0
/* getvbr():
 * Get the value of the vector base register.
 * Function is currently unused.
 */
unsigned long
getvbr(void)
{
	unsigned long value;

	STC(vbr,value);
	return(value);
}

/* getgbr();
 * Get the value of the global base register.
 * Function is currently unused.
 */
unsigned long
getgbr(void)
{
	unsigned long value;

	STC(gbr,value);
	return(value);
}
#endif

/* getsr():
 * Get the value of the status register.
 * This is needed for setting up the exception handling.
 * Used in "except_sh3.c".
 */
unsigned long
getsr(void)
{
	unsigned long value;

	STC(sr,value);
	return(value);
}

/* intsrestore():
 * Re-establish system interrupts here by using the status value
 * that was returned by an earlier call to intsoff().
 */
void
intsrestore(unsigned long sr)
{
	putsr(sr);
}

/* intsoff():
 * Return a value that can be used by intsrestore() (later) to 
 * restore the interrupt state.
 * When system interrupts are enabled, they must be disabled here.
 */
unsigned long
intsoff(void)
{
	unsigned long  sr;

	sr = getsr();
	putsr(sr | 0x00F0);
	return(sr);
}

